
from typing import List
from ceruleo.transformation.functional.transformerstep import TransformerStep
import pandas as pd
from functools import reduce 


class Sum(TransformerStep):
    """ 
    Concatenate multiple run-to-failure cycles vertically
    """
    def transform(self, X: List[pd.DataFrame]) -> pd.DataFrame:
        """
        Apply the concatenation 

        Parameters:
            X: List of run-to-failure cycles to concatenate

        Returns:
            A dataframe with the concatenated run-to-failure cycles
        """
        return reduce(lambda x, y: x.add(y, fill_value=0), X)


class Divide(TransformerStep):
    """
    Divide multiple run-to-failure cycles vertically
    """
    def transform(self, X: List[pd.DataFrame]) -> pd.DataFrame:
        """
        Apply the division

        Parameters:
            X: List of run-to-failure cycles to divide

        Returns:
            A dataframe with the divided run-to-failure cycles
        """
        return reduce(lambda x, y: x.divide(y, fill_value=0), X)



from math import sqrt

import numba
import numpy as np
import numpy.fft as fft
from numba import float32, jit, objmode
from numba.experimental import jitclass


@jit(nopython=True, error_model='numpy')
def get_moment(y: np.array, moment: int) -> np.float32:
    """
    Compute the (non centered) moment of the input distribution 

    Parameters:
        y: the discrete distribution from which one wants to calculate the moment
        moment: the moment one wants to calculate (choose 1,2,3, ... )
    
    Returns:
        The moment requested
    """
    return y.dot(np.arange(len(y), dtype=np.float64)**moment) / y.sum()


@jit(nopython=True, error_model='numpy')
def get_centroid(y: np.array) -> np.float32:
    """
    Compute the centroid of the input distribution (aka distribution mean, first moment)

    Parameters:
        y: the discrete distribution from which one wants to calculate the centroid

    Returns:
        The centroid of distribution y
    """
    return get_moment(y, 1)


@jit(nopython=True, error_model='numpy')
def get_variance(y: np.array) -> np.float32:
    """
    Compute the variance of the input distribution (aka distribution second central moment)

    Parameters:
        y: the discrete distribution from which one wants to calculate the variance

    Returns:
        The variance of distribution y
    """
    # Here we are implementing the formula var(X) = E[X**2] (second moment) - E[X]**2 (first moment squared)
    return get_moment(y, 2) - get_centroid(y) ** 2


@jit(nopython=True, error_model='numpy')
def get_skew(y: np.array) -> np.float32:
    """
    Calculates the skew as the third standardized moment. (Ref: https://en.wikipedia.org/wiki/Skewness#Definition)
    
    Parameters:
        y: the discrete distribution from which one wants to calculate the skew
    
    Returns:
        The skew of distribution y
    """

    variance = get_variance(y)
    # In the limit of a dirac delta, skew should be 0 and variance 0.  However, in the discrete limit,
    # the skew blows up as variance --> 0, hence return nan when variance is smaller than a resolution of 0.5:
    if variance < 0.5:
        return np.nan
    else:
        return (
            get_moment(y, 3) - 3 * get_centroid(y) *
            variance - get_centroid(y)**3
        ) / get_variance(y)**(1.5)


@jit(nopython=True, error_model='numpy')
def get_kurtosis(y: np.array) -> np.float32:
    """
    Calculates the kurtosis as the fourth standardized moment. (Ref: https://en.wikipedia.org/wiki/Kurtosis#Pearson_moments)
    
    Parameters:
        y: the discrete distribution from which one wants to calculate the kurtosis

    Returns:
        The kurtosis of distribution y
    """

    variance = get_variance(y)
    # In the limit of a dirac delta, kurtosis should be 3 and variance 0.  However, in the discrete limit,
    # the kurtosis blows up as variance --> 0, hence return nan when variance is smaller than a resolution of 0.5:
    if variance < 0.5:
        return np.nan
    else:
        return (
            get_moment(y, 4) - 4 * get_centroid(y) * get_moment(y, 3)
            + 6 * get_moment(y, 2) * get_centroid(y)**2 -
            3 * get_centroid(y)
        ) / get_variance(y)**2


@jit(debug=True, nopython=True, error_model='numpy')
def compute_frequency_features(x: np.array) -> np.array:
    """
    Returns the spectral centroid (mean), variance, skew, and kurtosis of the absolute fourier transform spectrum.
    
    Parameters:
        x: the time series on which the spectral features are calculated

    Returns:
        An array containig the spectral centroid, variance, skew, and kurtosis of the absolute fourier transform spectrum.
    """

    with objmode(fft_abs='float64[:]'):
        fft_abs = np.abs(fft.rfft(x))

    ps = fft_abs**2

    data = np.zeros(8, dtype=np.float32)
    data[0] = get_centroid(fft_abs)
    data[1] = get_variance(fft_abs)
    data[2] = get_skew(fft_abs)
    data[3] = get_kurtosis(fft_abs)
    data[4] = get_centroid(ps)
    data[5] = get_variance(ps)
    data[6] = get_skew(ps)
    data[7] = get_kurtosis(ps)
    return data



import logging
import zipfile
from pathlib import Path
from typing import List, Optional, Union

import numpy as np
import pandas as pd
from ceruleo import DATA_PATH
from ceruleo.dataset.ts_dataset import AbstractPDMDataset
from ceruleo.utils.download import download

logger = logging.getLogger(__name__)

DATASET_PATH = DATA_PATH / "C_MAPSS"


URL = "https://data.nasa.gov/download/ff5v-kuh6/application%2Fzip"



# Features used by
# Multiobjective Deep Belief Networks Ensemble for Remaining Useful Life Estimation in
# Prognostics Chong Zhang, Pin Lim, A. K. Qin,Senior Member, IEEE, and Kay Chen Tan,Fellow, IEEE
sensor_indices = np.array([2, 3, 4, 7, 8, 9, 11, 12, 13, 14, 15, 17, 20, 21]) + (4 - 1)

dependent_vars = ["RemainingUsefulLife"]
index_columns_names = ["UnitNumber", "Cycle"]
operational_settings_columns_names = ["OpSet" + str(i) for i in range(1, 4)]
sensor_measure_columns_names = ["SensorMeasure" + str(i) for i in range(1, 22)]
input_file_column_names = (
    index_columns_names
    + operational_settings_columns_names
    + sensor_measure_columns_names
)

operation_mode = {"FD001": 0, "FD002": 1, "FD003": 2, "FD004": 3}
engines = ["FD001", "FD002", "FD003", "FD004"]



        

def obtain_raw_files(raw_data_path: Path = DATASET_PATH, ):
    """Download and unzip the raw files

    Parameters:
        raw_data_path: Path where to store the dataset
    """
    raw_data_path = raw_data_path / "files"
    logger.info("Dataset not processed.")
    if not raw_data_path.is_dir():
        raw_data_path.mkdir(exist_ok=True, parents=True)
        ZIP_FILE = raw_data_path / "CMAPSSData.zip"
        if not ZIP_FILE.is_file():
            logger.info('Downloading file')            
            download(URL, ZIP_FILE)
        logger.info("Unzipping")
        with zipfile.ZipFile(ZIP_FILE, "r") as zip_ref:
            zip_ref.extractall(raw_data_path)
        logger.info("Removing zip file")
        ZIP_FILE.unlink()





def process_file_test(file):
    test_data = pd.read_csv(
        DATASET_PATH / "files" / ("test_" + file + ".txt"),
        names=input_file_column_names,
        delimiter=r"\s+",
        header=None,
    )
    truth_data = pd.read_csv(
        DATASET_PATH / "files" / ("RUL_" + file + ".txt"), delimiter=r"\s+", header=None
    )
    truth_data.columns = ["truth"]
    truth_data["UnitNumber"] = np.array(range(truth_data.shape[0])) + 1
    test_rul = test_data.groupby("UnitNumber")["Cycle"].max().reset_index()
    test_rul.columns = ["UnitNumber", "Elapsed"]
    test_rul = test_rul.merge(truth_data, on=["UnitNumber"], how="left")
    test_rul["Max"] = test_rul["Elapsed"] + test_rul["truth"]

    test_data = test_data.merge(test_rul, on=["UnitNumber"], how="left")
    test_data["RUL"] = test_data["Max"] - test_data["Cycle"]
    test_data.drop(["Max"], axis=1, inplace=True)
    return test_data


def prepare_train_data(data, factor: float = 0):
    df = data.copy()
    fdRUL = df.groupby("UnitNumber")["Cycle"].max().reset_index()
    fdRUL = pd.DataFrame(fdRUL)
    fdRUL.columns = ["UnitNumber", "max"]
    df = df.merge(fdRUL, on=["UnitNumber"], how="left")
    df["RUL"] = df["max"] - df["Cycle"]
    df.drop(columns=["max"], inplace=True)

    return df


def process_file_train(file):
    df = pd.read_csv(
        DATASET_PATH / "files"/ ("train_" + file + ".txt"),
        sep=r"\s+",
        names=input_file_column_names,
        header=None,
    )
    df = prepare_train_data(df)
    df["OpMode"] = operation_mode[file]
    return df


class CMAPSSDataset(AbstractPDMDataset):
    """C-MAPSS Dataset

    C-MAPSS stands for 'Commercial Modular Aero-Propulsion System Simulation' and it is a tool for the simulation 
    of realistic large commercial turbofan engine data. Each flight is a combination of a 
    series of flight conditions with a reasonable linear transition period to allow the 
    engine to change from one flight condition to the next. The flight conditions are arranged
    to cover a typical ascent from sea level to 35K ft and descent back down to sea level. 
     
    The fault was injected at a given time in one of the flights and persists throughout the 
    remaining flights, effectively increasing the age of the engine. The intent is to identify which 
    flight and when in the flight the fault occurred.

    [Dataset reference](https://data.nasa.gov/dataset/C-MAPSS-Aircraft-Engine-Simulator-Data/xaut-bemq)

    Available models are:
    
        - FD001
        - FD002
        - FD003
        - FD004


    Example:
        ```
        train_dataset = CMAPSSDataset(train=True, models='FD001')
        validation_dataset = CMAPSSDataset(train=False, models='FD001')
        ```

    Parameters:
        train: Weather to obtain the train data provided, by default True
        models: Names of the models, by default None (all models)
    """
    def __init__(
        self, train: bool = True, models: Optional[Union[str, List[str]]] = None
    ):
        super().__init__()
        obtain_raw_files(DATASET_PATH)
        if models is not None and isinstance(models, str):
            models = [models]
        self._validate_model_names(models)
        if train:
            processing_fun = process_file_train
        else:
            processing_fun = process_file_test
        self.lives = []

        for engine in engines:
            if models is not None and engine not in models:
                continue
            for _, g in processing_fun(engine).groupby("UnitNumber"):
                g.drop(columns=["UnitNumber"], inplace=True)
                g["Engine"] = engine
                self.lives.append(g)

    def _validate_model_names(self, models):
        if models is not None:
            for model in models:
                if model not in operation_mode:
                    raise ValueError(
                        f"Invalid model: valid model are {list(operation_mode.keys())}"
                    )

    def get_time_series(self, i):
        return self.lives[i]

    @property
    def n_time_series(self):
        return len(self.lives)

    @property
    def rul_column(self) -> str:
        return "RUL"



import scipy.stats
import pandas as pd
import numpy as np

def manual_expanding(df: pd.DataFrame, min_points:int= 1):
    to_compute = ['kurtosis', 'skewness', 'max', 'min', 'std', 'peak', 'impulse',
            'clearance', 'rms', 'shape', 'crest']
    dfs = []
    for c in df.columns:
        d = []
        for i in range(min_points-1):
            d.append([np.nan for f in to_compute])
        for end in range(min_points, df.shape[0]+1):
            data = df[c].iloc[:end]
            row = [manual_features(data, f) for f in to_compute]
            d.append(row)
        dfs.append(pd.DataFrame(d, columns=[f'{c}_{f}' for f in to_compute]))
    return pd.concat(dfs, axis=1)



def kurtosis(s: pd.Series) -> float:
    return scipy.stats.kurtosis(s.values, bias=False)


def skewness(s: pd.Series) -> float:
    return scipy.stats.skew(s.values, bias=False)


def max(s: pd.Series) -> float:
    return np.max(s.values)


def min(s: pd.Series) -> float:
    return np.min(s.values)


def std(s: pd.Series) -> float:
    return np.std(s.values, ddof=1)


def peak(s: pd.Series) -> float:
    return max(s) - min(s)


def impulse(s: pd.Series) -> float:
    return peak(s) / np.mean(np.abs(s))


def clearance(s: pd.Series) -> float:
    return peak(s) / (np.mean(np.sqrt(np.abs(s)))**2)


def rms(s: pd.Series) -> float:
    return np.sqrt(np.mean(s**2))


def shape(s: pd.Series) -> float:
    return rms(s) / np.mean(np.abs(s))


def crest(s: pd.Series) -> float:
    return peak(s) / rms(s)


feature_functions = {
    'crest': crest,
    'shape': shape,
    'rms': rms,
    'clearance': clearance,
    'impulse': impulse,
    'peak': peak,
    'std': std,
    'min': min,
    'max': max,
    'skewness': skewness,
    'kurtosis': kurtosis
}


def manual_features(s: pd.Series, name: str):
    return feature_functions[name](s)


from itertools import combinations
from typing import Dict, List, Optional, Tuple

import pandas as pd
from ceruleo.dataset.ts_dataset import AbstractPDMDataset
from ceruleo.dataset.utils import iterate_over_features
from pydantic import BaseModel

from ceruleo.utils import pydantic_to_dict


class CorrelationAnalysisElement(BaseModel):
    mean_correlation: float
    std_correlation: float
    max_correlation: float
    min_correlation: float
    abs_mean_correlation: float
    std_abs_mean_correlation: float


class CorrelationAnalysis(BaseModel):
    data: Dict[Tuple[str, str], CorrelationAnalysisElement]

    def get(self, feature_1: str, feature_2: str) -> CorrelationAnalysisElement:
        needle = (feature_1, feature_2)
        if needle not in self.data:
            needle = (feature_2, feature_1)

        if needle not in self.data:
            raise KeyError(f"Correlation between {feature_1} and {feature_2} not found")
        return self.data[needle]

    def to_pandas(self) -> pd.DataFrame:
        return (
            pd.DataFrame.from_dict(
                {(k[0], k[1]): pydantic_to_dict(v) for k, v in self.data.items()},
                orient="index",
            )
            .reset_index()
            .rename(columns={"level_0": "feature_1", "level_1": "feature_2"})
        )


def correlation_analysis(
    dataset: AbstractPDMDataset,
    features: Optional[List[str]] = None,
) -> CorrelationAnalysis:
    """
    Correlation Analysis
    Compute the correlation between all the features given an Iterable of executions.

    Parameters:
        dataset: Dataset of time series
        features: List of features to consider when computing the correlations

    Returns:
        A CorrelationAnalysis object with map indexed by two colun names and the following information:s

            - Mean Correlation
            - Std Correlation
            - Percentage of lives with a high correlation
            - Abs mean correlation
            - Std mean correlation
            - Max correlation
            - Min correlation

    """
    if features is None:
        features = sorted(list(dataset.common_features()))
    else:
        features = sorted(list(set(features).intersection(dataset.common_features())))
    features = dataset.get_features_of_life(0)[features].corr().columns
    correlated_features = []

    for ex in iterate_over_features(dataset):
        ex = ex[features]
        corr_m = ex.corr().fillna(0)

        correlated_features_for_execution = []

        for f1, f2 in combinations(features, 2):
            correlated_features_for_execution.append((f1, f2, corr_m.loc[f1, f2]))

        correlated_features.extend(correlated_features_for_execution)

    df = pd.DataFrame(correlated_features, columns=["Feature 1", "Feature 2", "Corr"])
    output = df.groupby(by=["Feature 1", "Feature 2"]).agg(
        {
            "Corr": [
                "mean",
                "std",
                "max",
                "min",
            ]
        }
    )

    # Calculate additional statistics
    output["Abs mean correlation"] = df.groupby(by=["Feature 1", "Feature 2"])[
        "Corr"
    ].apply(lambda x: x.abs().mean())
    output["Std abs mean correlation"] = df.groupby(by=["Feature 1", "Feature 2"])[
        "Corr"
    ].apply(lambda x: x.abs().std())

    output.columns = [
        "mean_correlation",
        "std_correlation",
        "max_correlation",
        "min_correlation",
        "abs_mean_correlation",
        "std_abs_mean_correlation",
    ]

    output = output.fillna(0)
    return CorrelationAnalysis(
        data={
            (k[0], k[1]): CorrelationAnalysisElement(
                mean_correlation=v["mean_correlation"],
                std_correlation=v["std_correlation"],
                max_correlation=v["max_correlation"],
                min_correlation=v["min_correlation"],
                abs_mean_correlation=v["abs_mean_correlation"],
                std_abs_mean_correlation=v["std_abs_mean_correlation"],
            )
            for k, v in output.iterrows()
        }
    )



from numba import guvectorize
import numpy as np
import pandas as pd

@guvectorize("float64[:], int64, int64, int64, float64[:]", "(m),(),(),()->()",
             cache=True, nopython=True)
def hurst_rs(x: np.array, min_chunksize: int, max_chunksize: int, num_chunksize: int, out: np.array) -> float:
    """
    Estimate the Hurst exponent using R/S method.

    Estimates the Hurst (H) exponent using the R/S method from the time series.
    The R/S method consists of dividing the series into pieces of equal size
    series_len and calculating the rescaled range. This repeats the process
    for several `series_len` values and adjusts data regression to obtain the H.
    `series_len` will take values between `min_chunksize` and `max_chunksize`,
    the step size from `min_chunksize` to `max_chunksize` can be controlled
    through the parameter `step_chunksize`.

    Parameters:
        x: A time series to calculate hurst exponent, must have more elements than `min_chunksize` and `max_chunksize`.
        min_chunksize: This parameter allow you control the minimum window size.
        max_chunksize: This parameter allow you control the maximum window size.
        num_chunksize: This parameter allow you control the size of the step from minimum to maximum window size. Bigger step means fewer calculations.
        out: One element array to store the output.
            

    Returns:
        Estimation of Hurst exponent.
        
    References
    ----------
        Hurst, H. E. (1951). Long term storage capacity of reservoirs. ASCE
        Transactions, 116(776), 770-808.
        Alessio, E., Carbone, A., Castelli, G. et al. Eur. Phys. J. B (2002) 27:
        197. http://dx.doi.org/10.1140/epjb/e20020150
    """
    N = len(x)
    max_chunksize += 1
    rs_tmp = np.empty(N, dtype=np.float64)
    chunk_size_list = np.linspace(min_chunksize, max_chunksize, num_chunksize)\
                        .astype(np.int64)
    rs_values_list = np.empty(num_chunksize, dtype=np.float64)

    # 1. The series is divided into chunks of chunk_size_list size
    for i in range(num_chunksize):
        chunk_size = chunk_size_list[i]

        # 2. it iterates on the indices of the first observation of each chunk
        number_of_chunks = int(len(x) / chunk_size)

        for idx in range(number_of_chunks):
            # next means no overlapping
            # convert index to index selection of each chunk
            ini = idx * chunk_size
            end = ini + chunk_size
            chunk = x[ini:end]

            # 2.1 Calculate the RS (chunk_size)
            z = np.cumsum(chunk - np.mean(chunk))
            rs_tmp[idx] = np.divide(
                np.max(z) - np.min(z),  # range
                np.nanstd(chunk)  # standar deviation
            )

        # 3. Average of RS(chunk_size)
        rs_values_list[i] = np.nanmean(rs_tmp[:idx + 1])

    # 4. calculate the Hurst exponent.
    H, c = np.linalg.lstsq(
        a=np.vstack((np.log(chunk_size_list), np.ones(num_chunksize))).T,
        b=np.log(rs_values_list)
    )[0]

    out[0] = H


def hurst_dma(prices: np.array, min_chunksize: int=8, max_chunksize: int=200, num_chunksize: int=5) -> float:
    """
    Estimate the Hurst exponent using the DMA method.

    Estimates the Hurst (H) exponent using the DMA method from the time series.
    The DMA method consists on calculate the moving average of size `series_len`
    and subtract it to the original series and calculating the standard
    deviation of that result. This repeats the process for several `series_len`
    values and adjusts data regression to obtain the H. `series_len` will take
    values between `min_chunksize` and `max_chunksize`, the step size from
    `min_chunksize` to `max_chunksize` can be controlled through the parameter
    `step_chunksize`.

    Parameters:
        prices: A time series to calculate hurst exponent, must have more elements than `min_chunksize` and `max_chunksize`.
        min_chunksize: This parameter allow you control the minimum window size.
        max_chunksize: This parameter allow you control the maximum window size.
        num_chunksize: This parameter allow you control the size of the step from minimum to maximum window size. Bigger step means fewer calculations.

    Returns:
        Estimation of hurst exponent.
       

    References
    ----------
        Alessio, E., Carbone, A., Castelli, G. et al. Eur. Phys. J. B (2002) 27:
        197. http://dx.doi.org/10.1140/epjb/e20020150

    """
    max_chunksize += 1
    N = len(prices)
    n_list = np.arange(min_chunksize, max_chunksize, num_chunksize, dtype=np.int64)
    dma_list = np.empty(len(n_list))
    factor = 1 / (N - max_chunksize)
    # sweeping n_list
    for i, n in enumerate(n_list):
        b = np.divide([n - 1] + (n - 1) * [-1], n)  # do the same as:  y - y_ma_n
        noise = np.power(signal.lfilter(b, 1, prices)[max_chunksize:], 2)
        dma_list[i] = np.sqrt(factor * np.sum(noise))

    H, const = np.linalg.lstsq(
        a=np.vstack([np.log10(n_list), np.ones(len(n_list))]).T,
        b=np.log10(dma_list)
    )[0]
    return H


def hurst_dsod(x: np.array) -> float:
    """
    Estimate Hurst exponent on data timeseries using the DSOD (Discrete Second Order Derivative).

    The estimation is based on the discrete second order derivative. Consists on
    get two different noise of the original series and calculate the standard
    deviation and calculate the slope of two point with that values.
    source: https://gist.github.com/wmvanvliet/d883c3fe1402c7ced6fc

    Parameters:
        x: Time series to estimate the Hurst exponent for.
        

    Returns:
       Estimation of the Hurst exponent for the given time series.
    
    References
    ----------
        Istas, J.; G. Lang (1994), “Quadratic variations and estimation of the local
        Hölder index of data Gaussian process,” Ann. Inst. Poincaré, 33, pp. 407–436.


    Notes
    ----------
        This hurst_ets is data literal traduction of wfbmesti.m of waveleet toolbox from matlab.
    """
    y = np.cumsum(np.diff(x, axis=0), axis=0)

    # second order derivative
    b1 = [1, -2, 1]
    y1 = signal.lfilter(b1, 1, y, axis=0)
    y1 = y1[len(b1) - 1:]  # first values contain filter artifacts

    # wider second order derivative
    b2 = [1,  0, -2, 0, 1]
    y2 = signal.lfilter(b2, 1, y, axis=0)
    y2 = y2[len(b2) - 1:]  # first values contain filter artifacts

    s1 = np.mean(y1 ** 2, axis=0)
    s2 = np.mean(y2 ** 2, axis=0)

    return 0.5 * np.log2(s2 / s1)


def hurst_exponent(prices: Option[np.ndarray, pd.Series, pd.DataFrame], min_chunksize: int =8, max_chunksize: int =200, num_chunksize: int =5,
                   method: str ='RS') -> float:
    """
    Estimates Hurst Exponent.

    Estimate the hurst exponent following one of 3 methods. Each method

    Parameters:
        prices: A time series to estimate hurst exponent.
        min_chunksize : int, optional
            Minimum chunk  size of the original series. This parameter doesn't have any effect with DSOD method, by default 8
        max_chunksize : int, optional
            Maximum chunk size of the original series. This parameter doesn't have any effect with DSOD method. by default 200.
        num_chunksize : int, optional
            Step used to select next the chunk size which divide the original series. This parameter doesn't have any effect with DSOD method, by default 5.
        method : {'RS', 'DMA', 'DSOD', 'all'}. Valid values are: RS : rescaled range, DMA : deviation moving average, DSOD : discrete second order derivative.


    Returns:
        Estimation of hurst_exponent according to the method selected.
        

    References
    ----------
        RS: Hurst, H. E. (1951). Long term storage capacity of reservoirs. ASCE
            Transactions, 116(776), 770-808.
        DMA: Alessio, E., Carbone, A., Castelli, G. et al. Eur. Phys. J. B (2002)
            27: 197. http://dx.doi.org/10.1140/epjb/e20020150
        DSOD: Istas, J.; G. Lang (1994), “Quadratic variations and estimation of
            the local Hölder index of data Gaussian process,” Ann. Inst. Poincaré,
            33, pp. 407–436.

    Notes
    ----------
        The hurst exponent is an estimation which is important because there is no
        data closed equation for it instead we have some methods to estimate it with
        high variations among them.

    See Also
    ----------
        hurst_rs, hurst_dma, hurst_dsod
    """
    if len(prices) == 0:
        return np.nan
    # extract array
    arr = prices.__array__()
    # choose data method
    if method == 'RS':
        if prices.ndim > 1:
            h = hurst_rs(np.diff(arr, axis=0).T, min_chunksize, max_chunksize,
                         num_chunksize)
        else:
            h = hurst_rs(np.diff(arr), min_chunksize, max_chunksize,
                         num_chunksize)
    elif method == 'DMA':
        h = hurst_dma(arr, min_chunksize, max_chunksize, num_chunksize)
    elif method == 'DSOD':
        h = hurst_dsod(arr)
    elif method == 'all':
        return [
            hurst_exponent(arr, min_chunksize, max_chunksize, num_chunksize, 'RS'),
            hurst_exponent(arr, min_chunksize, max_chunksize, num_chunksize, 'DMA'),
            hurst_exponent(arr, min_chunksize, max_chunksize, num_chunksize, 'DSOD')
        ]
    else:
        raise NotImplementedError('The method choose is not implemented.')
    
    return h


from typing import Tuple

import numpy as np
import tensorflow as tf
from tensorflow.keras import Input, Model
from tensorflow.keras.layers import (Activation, Add, BatchNormalization,
                                     Concatenate, Conv1D, Dense,
                                     GlobalAveragePooling1D, MaxPool1D)


def _inception_module(
    input_tensor,
    use_bottleneck: bool,
    bottleneck_size: int,
    kernel_size: int,
    inception_number: int,
    nb_filters: int,
    stride=1,
    activation="linear",
):

    if use_bottleneck and int(input_tensor.shape[-1]) > 1:
        input_inception = Conv1D(
            filters=bottleneck_size,
            kernel_size=1,
            padding="same",
            activation=activation,
            use_bias=False,
        )(input_tensor)
    else:
        input_inception = input_tensor

    kernel_size_s = [kernel_size // (2**i) for i in range(inception_number)]

    conv_list = []

    for i in range(len(kernel_size_s)):
        conv_list.append(
            Conv1D(
                filters=nb_filters,
                kernel_size=kernel_size_s[i],
                strides=stride,
                padding="same",
                activation=activation,
                use_bias=False,
            )(input_inception)
        )

    max_pool_1 = MaxPool1D(pool_size=3, strides=stride, padding="same")(input_tensor)

    conv_6 = Conv1D(
        filters=nb_filters,
        kernel_size=1,
        padding="same",
        activation=activation,
        use_bias=False,
    )(max_pool_1)

    conv_list.append(conv_6)

    x = Concatenate(axis=2)(conv_list)
    x = BatchNormalization()(x)
    x = Activation(activation="relu")(x)
    return x


def _shortcut_layer(input_tensor, out_tensor):
    shortcut_y = Conv1D(
        filters=int(out_tensor.shape[-1]),
        kernel_size=1,
        padding="same",
        use_bias=False,
    )(input_tensor)
    shortcut_y = BatchNormalization()(shortcut_y)

    x = Add()([shortcut_y, out_tensor])
    x = Activation("relu")(x)
    return x


def InceptionTime(
    input_shape: Tuple[int, int],
    *,
    nb_filters: int = 32,
    use_residual: bool = True,
    use_bottleneck: bool = True,
    depth: int = 6,
    kernel_size: int = 41,
    bottleneck_size: int = 32,
    inception_number: int = 3,
) -> tf.keras.Model:
    """InceptionTime


    Ismail Fawaz, H., Lucas, B., Forestier, G., Pelletier, C., Schmidt, D. F., Weber, J., ... & Petitjean, F. (2020).
    Inceptiontime: Finding alexnet for time series classification.
    Data Mining and Knowledge Discovery, 34(6), 1936-1962.

    [Reference](https://link.springer.com/article/10.1007/s10618-020-00710-y)


    Parameters:

        input_shape: Input shape
        nb_filters: number of fiters
        use_residual: Wether to use residual connections
        use_bottleneck : bool, optional
            _description_, by default True
        depth: max_depth
        kernel_size: kernel size
        bottleneck_size: bottleneck size
        inception_number: iception number

    Returns:

        model: Model
    """

    input = Input(shape=input_shape)

    x = input
    input_res = input

    for d in range(depth):

        x = _inception_module(
            x,
            use_bottleneck,
            bottleneck_size,
            kernel_size,
            inception_number,
            nb_filters,
        )

        if use_residual and d % 3 == 2:
            x = _shortcut_layer(input_res, x)
            input_res = x

    gap_layer = GlobalAveragePooling1D()(x)

    output_layer = Dense(1, activation="relu")(gap_layer)

    model = Model(inputs=input, outputs=output_layer)

    return model




from ceruleo.dataset.analysis.sample_rate import sample_rate

from ceruleo.dataset.analysis.correlation import correlation_analysis
from ceruleo.dataset.analysis.distribution import features_divergeces


from copy import copy
from typing import Union


def root_nodes(step_or_pipe: Union["TemporisPipeline", "TransformerStep"]):
    from ceruleo.transformation.functional.pipeline.pipeline import Pipeline

    if isinstance(step_or_pipe, Pipeline):
        final_step = step_or_pipe.final_step
    else:
        final_step = step_or_pipe
    final_step = final_step
    visited = set([final_step])
    to_process = copy(final_step.previous)

    while len(to_process) > 0:
        t = to_process.pop()
        if t not in visited:
            visited.add(t)
            to_process.extend(t.previous)

    return [n for n in visited if len(n.previous) == 0]


def dfs_iterator(step_or_pipe: Union["TemporisPipeline", "TransformerStep"]):

    from ceruleo.transformation.functional.pipeline.pipeline import Pipeline

    if isinstance(step_or_pipe, Pipeline):
        final_step = step_or_pipe.final_step
    else:
        final_step = step_or_pipe
    final_step = final_step
    visited = set([])
    Q = copy(root_nodes(final_step))
    while len(Q) > 0:
        node = Q.pop()
        if node in visited:
            continue
        visited.add(node)

        yield node

        if len(node.next) == 0:
            continue
        for n in node.next:
            Q.append(n)


class topological_sort_iterator:
    def __init__(self, step_or_pipe: Union["TemporisPipeline", "TransformerStep"]):
        from ceruleo.transformation.functional.pipeline.pipeline import Pipeline

        if isinstance(step_or_pipe, Pipeline):
            final_step = step_or_pipe.final_step
        else:
            final_step = step_or_pipe
        self.final_step = final_step
        self.in_degree = {}

    def _initialize_degrees(self):
        for node in dfs_iterator(self.final_step):
            if node not in self.in_degree:
                self.in_degree[node] = 0
            if len(node.next) == 0:
                continue
            for n in node.next:
                if n not in self.in_degree:
                    self.in_degree[n] = 0
                self.in_degree[n] += 1

    def __iter__(self):
        self.Q = root_nodes(self.final_step)
        self._initialize_degrees()
        self.current_node = None
        self.traversed = []
        return self

    def graph_updated(self):
        self.in_degree = {}
        self._initialize_degrees()
        for node in self.traversed:
            for n in node.next:
                self.in_degree[n] -= 1

    def __next__(self):
        if self.current_node is not None and len(self.current_node.next) > 0:

            for n in self.current_node.next:
                if n not in self.in_degree:
                    self.graph_updated()
                    break
                else:
                    self.in_degree[n] -= 1
            for n in self.current_node.next:
                if self.in_degree[n] == 0:
                    self.Q.append(n)
        if len(self.Q) == 0:
            raise StopIteration
        self.current_node = self.Q.pop(0)
        self.traversed.append(self.current_node)

        return self.current_node


def nodes(pipe : "TemporisPipeline"):
    _nodes = set()
    for root in root_nodes(pipe):        
        for node in topological_sort_iterator(root):
            _nodes.add(node)
    return list(_nodes)


def edges(pipe: "TemporisPipeline"):
    _edges = set()
    for n in nodes(pipe):
        for n_next in n.next:
            if (n, n_next) not in _edges and (n_next, n) not in _edges:
                _edges.add((n, n_next))
    return _edges


from typing import Optional

import numpy as np
import pandas as pd
from scipy.signal import savgol_filter
from sklearn.cluster import MiniBatchKMeans
from ceruleo.transformation import TransformerStep


class SavitzkyGolayTransformer(TransformerStep):
    """Filter each feature using LOESS

    Parameters:
        window: window size of the filter
        order:  Order of the filter, by default 2
        name: Step name
    """

    def __init__(self, window: int, order: int = 2, name: Optional[str] = None):
        super().__init__(name=name)
        self.window = window
        self.order = order

    def transform(self, X: pd.DataFrame, y=None) -> pd.DataFrame:
        """
        Return a new dataframe with the features filtered

        Parameters:
            X: Input life

        Returns:
            A new DatafFrame with the same index as the input with the features filtered
        """
        if X.shape[0] > self.window:
            return pd.DataFrame(
                savgol_filter(X, self.window, self.order, axis=0),
                columns=X.columns,
                index=X.index,
            )
        else:
            return X


class MeanFilter(TransformerStep):
    """
    Filter each feature using a rolling mean filter

    Parameters:
        window: Size of the rolling window
        min_periods: Minimum number of non-null points of the rolling window, by default 15
        name: Name of the step, by default None
        center: Wether the guassian window should be centered, by default False
    """

    def __init__(
        self,
        window: int,
        center=True,
        min_periods: int = 15,
        name: Optional[str] = None,
    ):
        super().__init__(name=name)
        self.window = window
        self.min_periods = min_periods
        self.center = center

    def transform(self, X: pd.DataFrame, y=None) -> pd.DataFrame:
        """
        Return a new dataframe with the features filtered

        Parameters:
            X: Input life

        Returns:
            A new DatafFrame with the same index as the input with the features filtered
        """
        return X.rolling(
            self.window,
            min_periods=self.min_periods,
            center=self.center,
        ).mean(numeric_only=True)


class MedianFilter(TransformerStep):
    """Filter each feature using a rolling median filter

    Parameters:
        window: Size of the rolling window
        min_periods: Minimum number of points of the rolling window, by default 15
        name: Name of the step, by default None
    """

    def __init__(self, window: int, min_periods: int = 15, name: Optional[str] = None):
        super().__init__(name=name)
        self.window = window
        self.min_periods = min_periods

    def transform(self, X: pd.DataFrame, y=None) -> pd.DataFrame:
        """
        Return a new dataframe with the features filtered

        Parameters:
            X: Input life

        Returns:
            A new DatafFrame with the same index as the input with the features filtered
        """
        return X.rolling(self.window, min_periods=self.min_periods).median(
            numeric_only=True
        )


class OneDimensionalKMeans(TransformerStep):
    """
    Clusterize each feature into a number of clusters

    Parameters:
        n_clusters: Number of clusters, by default 5

    """

    def __init__(self, n_clusters: int = 5, name: Optional[str] = None):
        super().__init__(name=name)
        self.clusters = {}
        self.n_clusters = n_clusters

    def partial_fit(self, X: pd.DataFrame):
        """
        Fit the model to the input data to obtain the clusters

        Parameters:
            X: Input life
        """
        if len(self.clusters) == 0:
            for c in X.columns:
                self.clusters[c] = MiniBatchKMeans(
                    n_clusters=self.n_clusters, n_init="auto"
                )

        for c in X.columns:
            self.clusters[c].partial_fit(np.atleast_2d(X[c]).T)
        return self

    def transform(self, X: pd.DataFrame, y=None) -> pd.DataFrame:
        """
        Transform the input dataframe

        Parameters:
            X: Input life

        Returns:
            A new DataFrame with the same index as the input. Each feature is replaced with the clusters of each point
        """
        X = X.copy()
        for c in X.columns:
            X[c] = self.clusters[c].cluster_centers_[
                self.clusters[c].predict(np.atleast_2d(X[c]).T)
            ]
        return X


class MultiDimensionalKMeans(TransformerStep):
    """
    Clusterize data points and replace each feature with the centroid feature it belongs to

    Parameters:
        n_clusters: Number of clusters to obtain by default 5
        name: Name of the step, by default None
    """

    def __init__(self, n_clusters: int = 5, name: Optional[str] = None):
        super().__init__(name=name)
        self.n_clusters = n_clusters
        self.clusters = MiniBatchKMeans(n_clusters=self.n_clusters, n_init="auto")

    def partial_fit(self, X: pd.DataFrame):
        """
        Fit the model to the input data to obtain the clusters

        Parameters:
            X: Input life
        """
        self.clusters.partial_fit(X)
        return self

    def transform(self, X: pd.DataFrame, y=None) -> pd.DataFrame:
        """Transform the input life with the centroid information

        Parameters:
            X: Input life

        Returns:
            A new DataFrame in which each point was replaced by the centroid it belongs to
        """

        X = X.copy()
        X[:] = self.clusters.cluster_centers_[self.clusters.predict(X)]
        return X


class EWMAFilter(TransformerStep):
    """
    Filter each feature using EWM (Exponential Moving Window)

    Parameters:
        span: Time constant of the EMA (Exponential Moving Average)
        name: Name of the step, by default None
    """

    def __init__(self, span: float, name: Optional[str] = None):
        super().__init__(name=name)
        self.span = span

    def transform(self, X: pd.DataFrame, y=None) -> pd.DataFrame:
        """
        Return a new dataframe with the features filtered

        Parameters:
            X: Input life

        Returns:
            A new DatafFrame with the same index as the input with the features filtered
        """
        return X.ewm(span=self.span, ignore_na=True).mean()


class GaussianFilter(TransformerStep):
    """
    Apply a gaussian filter

    Parameters:
        window_size: Size of the gaussian filter
        std: Standard deviation of the filter
        min_points: Minimun nomber of points of the rolling window, by default 1
        center: Wether the guassian window should be centered, by default False
    """

    def __init__(
        self,
        window_size: int,
        std: float,
        min_points: int = 1,
        center: bool = False,
        *args,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.window_size = window_size
        self.std = std
        self.center = center
        self.min_points = min_points

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Return a new dataframe with the features filtered

        Parameters:
            X: Input life

        Returns:
            A new DatafFrame with the same index as the input with the features filtered
        """
        return X.rolling(
            window=self.window_size,
            win_type="gaussian",
            center=self.center,
            min_periods=self.min_points,
        ).mean(std=self.std)




from typing import Any, Callable, Union

import numpy as np
import pandas as pd


class AbstractSampleWeights:
    """
    The base class for the sample weight provider
    """

    def __call__(self, y: Union[np.ndarray, pd.DataFrame], i: int, metadata):
        raise NotImplementedError



def get_value(y: Union[np.ndarray, pd.DataFrame], i:int) -> float:
    if isinstance(y, np.ndarray):
        if len(y.shape) > 1:
            return y[i, 0]
        else:
            return  y[i]
    elif isinstance(y, pd.DataFrame):
        return y.iloc[i, 0] 
    elif isinstance(y, pd.Series):
        return y.iloc[i]
    else:
        raise ValueError(f"Unsupported type {type(y)}")
   


class NotWeighted(AbstractSampleWeights):
    """
    Simplest sample weight provider

    Provide 1 as a sample weight for every sample
    """

    def __call__(self, y: Union[np.ndarray, pd.DataFrame], i: int, metadata):
        return 1


"""
The Sample Weight type is a callable with the following signature

    fun(y, i:int, metadata)

Given the target and the sample index `i` it returns the sample weight for sample `i`
"""
SampleWeight = Union[AbstractSampleWeights, Callable[[np.ndarray, int, Any], float]]


class RULInverseWeighted(AbstractSampleWeights):
    """
    Weight each sample by the inverse of the RUL
    """

    def __call__(self, y : Union[np.ndarray, pd.DataFrame], i: int, metadata):
        return 1 / (get_value(y, i) + 1)

        


class InverseToLengthWeighted(AbstractSampleWeights):
    """
    Weights samples according to the duration of the run-to-failure cycle they belong to.

    All points in the run-to-cycle are weighted equally inverse to the cycle duration

    """

    def __call__(self, y:Union[np.ndarray, pd.DataFrame], i: int, metadata):
        return 1 / get_value(y, 0)


class ExponentialDecay(AbstractSampleWeights):
    """
    Weight samples with an exponential decay function based on the RUL.

    """

    def __init__(self, *, near_0_at: float):
        
        super().__init__()
        self.alpha = -((near_0_at) ** 2) / np.log(0.000001)

    def __call__(self, y:Union[np.ndarray, pd.DataFrame], i: int, metadata):
        return ( np.exp(-(get_value(y,i) ** 2) / self.alpha))



from functools import partial
from typing import List, Optional

import numpy as np
import pandas as pd
from ceruleo.transformation import TransformerStep
from pandas.api.types import is_timedelta64_dtype
from scipy.stats import norm


class TargetToClasses(TransformerStep):
    """Transform the RUL values into a discrete set of classes

    Parameters
        bins: Bins limits
    """

    def __init__(self, bins: List[float]):
        super().__init__()
        self.bins = bins

    def transform(self, X):
        X_new = pd.DataFrame(index=X.index)
        X_new["RUL_class"] = np.digitize(X.iloc[:, 0].values, self.bins, right=False)
        return X_new


class PicewiseRUL(TransformerStep):
    """Transform the RUL clipping the maximum value

    Parameters

        max_life:  Maximum threshold for clipping the RUL values
        name:  Name of the step, by default None
    """

    def __init__(self, *, max_life: float = np.inf, name: Optional[str] = None):
        super().__init__(name=name)
        self.max_life = max_life

    def transform(self, X: pd.DataFrame):
        return X.clip(0, self.max_life)


class PicewiseRULQuantile(PicewiseRUL):
    """Clip the RUL  values using a quantile value as threshold

    Parameters

        quantile: Value between 0 and 1 to compute the RUL threshold
        name: Name of the step
    """

    def __init__(self, quantile: float, name: Optional[str] = None):
        super().__init__(name)
        self.quantile = quantile

    def fit(self, X, y=None):
        self.max_life = np.quantile(X, self.quantile)
        return self


class RULBinarizer(TransformerStep):
    """Convert the RUL target into a binary vector

    Useful for determining point of failures.


    Parameters
        t: Starting point of failure
        Every sample with a RUL greater than t will be 1 and the rest 0
    """

    def __init__(self, t: float, **kwargs):
        super().__init__(**kwargs)
        self.t = t

    def transform(self, X):
        return (X > self.t).astype("int")



import numpy as np 
from ceruleo.results.results import split_lives_indices
import pytest

class TestResults:
    def test_1(self):

        v1 = np.linspace(25, 0, 50)
        v2 = np.linspace(17, 5, 45)
        v3 = np.linspace(50, 7, 50)  
        y_true = np.hstack((v1,v2,v3))
        indices = split_lives_indices(y_true)
        assert (y_true[indices[0]] == v1).all()
        assert (y_true[indices[1]] == v2).all()
        assert (y_true[indices[2]] == v3).all()
        assert list(map(len, indices)) == [50, 45, 50]

        y_true = v1
        indices = split_lives_indices(y_true)
        assert (y_true[indices[0]] == v1).all()
        assert list(map(len, indices)) == [50]

        with pytest.raises(AssertionError):
            y_true = np.linspace(25, 0, 1)
            indices = split_lives_indices(y_true)
            assert (y_true[indices[0]] == v1).all()
            assert list(map(len, indices)) == []




"""
The dataset package provides an abstract class to handle dataset for predictive maintenance. 
The constitutive unit of the dataset are run-to-failure cycles, stored as a DataFrame. The dataset
should provides an access to each cycle separately.

Models implemented in this library accepts as input method for fitting and predicting,
an instance of a Dataset defined in this package.
"""


from abc import ABC, abstractmethod
import pandas as pd 
import numpy as np

class RULColumn(ABC):
    @abstractmethod
    def get(self, data: pd.DataFrame) -> np.ndarray:
        raise NotImplementedError
    
class NumberOfRowsRULColumn(RULColumn):
    def get(self, data: pd.DataFrame) -> np.ndarray:
        return np.arange(data.shape[0] - 1, -1, -1)
    
class CycleRULColumn(RULColumn):
    def __init__(self, cycle_column: str):
        self.cycle_column = cycle_column
    
    def get(self, data: pd.DataFrame) -> np.ndarray:
        return data[self.cycle_column].max() - data[self.cycle_column]
    
class DatetimeRULColumn(RULColumn):
    datetime_column: str
    units: str

    def __init__(self, datetime_column: str, units: str = "h"):
        self.datetime_column = datetime_column
        self.units = units
    
    def get(self, data: pd.DataFrame) -> pd.Series:
        q =  (data[self.datetime_column].max() - data[self.datetime_column]).dt.total_seconds() 
        if self.units == "d":
            return q / (3600*24)
        if self.units == "h":
            return q / 3600
        elif self.units == "min":
            return q / 60
        elif self.units == "s":
            return q
        elif self.units == "ms":
            return q * 1000
        


    


import numpy as np 
from typing import Callable

# Window len = L, Stride len/stepsize = S
def _strided_app(a: np.array, L: int, S: int) -> np.array:  
    """
    Returns an array that is strided

    Parameters:
        a: Array to be strided
        L: Length of the window
        S: Stride (S=L/stepsize)
    """
    nrows = ((a.size-L)//S)+1
    n = a.strides[0]
    r =  np.lib.stride_tricks.as_strided(
        a, 
        shape=(nrows, L), 
        strides=(S*n, n),
        writeable=False)
    last = r.shape[0]*r.shape[1]
    l = r.tolist()
    if last < len(a):
        l.append(a[last:])
    return l
    

def apply_rolling_data(values : np.ndarray, function: Callable[[np.ndarray], np.ndarray], window: int, step: int =1) -> np.array:
    """
    Perform a rolling window analysis at the column `col` from `data`

    Given a dataframe `data` with time series, call `function` at sections of length `window` at the data of column `col`. Append the results to `data` at a new columns with name `label`.

    Parameters:
        values: 1-D Time series of data
        function: Function to be called to calculate the rolling window analysis, the function must receive as input an array or pandas series. Its output must be either a number or a pandas series
        window: Length of the window to perform the analysis
        step: Step to take between two consecutive windows, by default 1

    Returns:
        data: Columns generated by the function applied
    """

    x = _strided_app(values, window, step)

    return np.vstack([function(np.array(b)) for b in x])
    



from typing import List

import numpy as np
import pandas as pd
import scipy.sparse as sparse
from ceruleo.transformation.functional.mixin import TransformerStepMixin



class Concatenate(TransformerStepMixin):
    """Join two transformers

    Parameters
    ----------
    add_prefix : bool, optional
        Whether to add prefix , by default True
    """
    def __init__(self, *, add_prefix:bool= True):

        super().__init__()
        self.add_prefix = add_prefix 


    def merge_dataframes_by_column(self, Xs):
        # indices = [X.index for X in Xs]
        # TODO: Check equal indices
        names = [n.name for n in self.previous]
        def new_feature_name(otherX, name, i):
            if self.add_prefix:
                return otherX.add_prefix(f'{name}_{i}_')
            else:
                return otherX
        X = pd.concat(
            [new_feature_name(otherX, name, i) for i, (name, otherX) in enumerate(zip(names, Xs))], axis=1
        )
        return X

    def transform(self, Xs: List[pd.DataFrame]):
        if not Xs:
            # All transformers are None
            return np.zeros((Xs[0].shape[0], 0))
        if any(sparse.issparse(f) for f in Xs):
            Xs = sparse.hstack(Xs).tocsr()
        else:
            Xs = self.merge_dataframes_by_column(Xs)
        return Xs

    def get_params(self, deep=True):
        params = super().get_params(deep=deep)
        return params



from typing import Iterable, List, Union

import pandas as pd
from ceruleo.dataset.ts_dataset import AbstractPDMDataset
from ceruleo.transformation.functional.graph_utils import (
    dfs_iterator,
    topological_sort_iterator,
)
from ceruleo.transformation.functional.pipeline.cache_store import CacheStoreType
from ceruleo.transformation.functional.pipeline.runner import CachedPipelineRunner
from ceruleo.transformation.functional.transformerstep import TransformerStep
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.base import clone as sklearn_clone

class Pipeline(BaseEstimator, TransformerMixin):
    """Transformation pipeline

    Parameters:
        final_step: The final step of the transformation
        cache_type: Cache storage mode
    """

    final_step: TransformerStep
    fitted_: bool
    cache_type: CacheStoreType
    runner: CachedPipelineRunner

    def __init__(self, final_step, cache_type: CacheStoreType = CacheStoreType.MEMORY):
        self.final_step = final_step
        self.fitted_ = False
        self.cache_type = cache_type
        self.runner = CachedPipelineRunner(final_step, cache_type)

    def find_node(
        self, name: str
    ) -> Union[List[TransformerStep], TransformerStep, None]:
        """Find a transformation node given a name

        Parameters:
            name: Name of the step to find

        Returns:

            steps: Steps located in the pipeline.

        """
        matches = []
        for node in dfs_iterator(self.final_step):
            if node.name == name:
                matches.append(node)

        if len(matches) == 1:
            return matches[0]
        elif len(matches) > 1:
            return matches
        else:
            return None

    def fit(
        self,
        dataset: Union[AbstractPDMDataset, pd.DataFrame],
        show_progress: bool = False,
    ):
        """Fit a pipeline using a dataset

        The CachedPipelineRunner is called to fit

        Parameters:

            dataset: A dataset of a run-to-failure cycle
            show_progress: Wether to show the progress when fitting

        Returns:
            s : Pipeline
        """
        if isinstance(dataset, pd.DataFrame):
            dataset = [dataset]
        c = self.runner.fit(dataset, show_progress=show_progress)
        self.column_names = c.columns
        self.fitted_ = True

        return self

    def partial_fit(
        self,
        dataset: Union[AbstractPDMDataset, pd.DataFrame],
        show_progress: bool = False,
    ):
        self.fit(dataset, show_progress=show_progress)

    def transform(self, df: Union[pd.DataFrame, Iterable[pd.DataFrame]]):
        """Transform a run-to-cycle failure or a dataset

        The CachedPipelineRunner is called to transform

        Parameters:

            df: A dataset of a run-to-failure cycle

        Returns:
            s : list of data frames
        """
        return self.runner.transform(df)

    def description(self):
        data = []
        for node in topological_sort_iterator(self):
            data.append(node.description())
        return data

    def get_params(self, deep: bool = False):
        params = {"cache_type": self.cache_type, "final_step": self.final_step}
        if deep:
            for node in topological_sort_iterator(self):
                p = node.get_params(deep)
                for k in p.keys():
                    params[f"{node.name}__{k}"] = p[k]
        return params
    
    def __sklearn_clone__(self):
        g = {node: sklearn_clone(node) for node  in dfs_iterator(self.final_step)}
        for k in g.keys():
            g[k].clear_connections()
        for node in dfs_iterator(self.final_step):
            for next_node in node.next:
                g[node].add_next(g[next_node])

        return Pipeline(
            final_step=g[self.final_step],
            cache_type=self.cache_type
        )


def make_pipeline(
    *steps, cache_type: CacheStoreType = CacheStoreType.MEMORY
) -> Pipeline:
    """Build a pipeline

    Example:

        make_pipeline(
            ByNameFeatureSelector(features=FEATURES),
            Clip(lower=-2, upper=2),
            IndexMeanResampler(rule='500s')
        )

    Parameters:

        steps: List of steps
        cache_type: Where to store the pipeline intermediate steps

    Returns:

        TemporisPipeline: The created pipeline
    """
    step = steps[0]
    for next_step in steps[1:]:
        step = next_step(step)

    return Pipeline(step, cache_type=cache_type)



import gzip
import io
import logging
import os
import pickle
import shutil
import tarfile
from enum import Enum
from pathlib import Path
from typing import List, Optional, Union

import gdown
import numpy as np
import pandas as pd
from tqdm.auto import tqdm

from ceruleo import CACHE_PATH, DATA_PATH
from ceruleo.dataset.builder.builder import DatasetBuilder
from ceruleo.dataset.builder.cycles_splitter import FailureDataCycleSplitter
from ceruleo.dataset.builder.output import DatasetFormat, LocalStorageOutputMode
from ceruleo.dataset.builder.rul_column import NumberOfRowsRULColumn
from ceruleo.dataset.ts_dataset import AbstractPDMDataset, PDMDataset

logger = logging.getLogger(__name__)

COMPRESSED_FILE = "phm_data_challenge_2018.tar.gz"
FOLDER = "phm_data_challenge_2018"


URL = "https://drive.google.com/uc?id=15Jx9Scq9FqpIGn8jbAQB_lcHSXvIoPzb"
OUTPUT = COMPRESSED_FILE


def download(url: str, path: Path):
    logger.info("Downloading dataset...")
    gdown.download(url, str(path / OUTPUT), quiet=False)


class FailureType(Enum):
    """Failure types availables for the dataset.

    Possible values are:
    ```
    FailureType.FlowCoolPressureDroppedBelowLimit
    FailureType.FlowcoolPressureTooHighCheckFlowcoolPump
    FailureType.FlowcoolLeak
    ```
    """

    FlowCoolPressureDroppedBelowLimit = "FlowCool Pressure Dropped Below Limit"
    FlowcoolPressureTooHighCheckFlowcoolPump = (
        'Flowcool Pressure Too High Check Flowcool Pump'
    )
    FlowcoolLeak = "Flowcool leak"
    FlowcoolPressureTooHighCheckFlowcoolPumpNoWaferID = 'Flowcool Pressure Too High Check Flowcool Pump [NoWaferID]'


    @staticmethod
    def that_starth_with(s: str):
        for f in FailureType:
            if s.startswith(f.value):
                return f
        return None


class PHMDataset2018(PDMDataset):
    """PHM 2018 Dataset

    The 2018 PHM dataset is a public dataset released by Seagate which contains the execution of 20 different
    ion milling machines. They distinguish three different failure causes and provide 22 features,
    including user-defined variables and sensors.

    Three faults are present in the dataset

    * Fault mode 1 occurs when flow-cool pressure drops.
    * Fault mode 2 occurs when flow-cool pressure becomes too high.
    * Fault mode 3 represents flow-cool leakage.

    [Dataset reference](https://phmsociety.org/conference/annual-conference-of-the-phm-society/annual-conference-of-the-prognostics-and-health-management-society-2018-b/phm-data-challenge-6/)

    Example:

    ```py
    dataset = PHMDataset2018(
        failure_types=FailureType.FlowCoolPressureDroppedBelowLimit,
        tools=['01_M02']
    )
    ```



    Parameters:
        path: Path where the dataset is located
    """

    failure_types: Optional[List[FailureType]]
    tools: Optional[List[str]]

    def __init__(
        self,
        path: Path = DATA_PATH,
        url: str = URL,
        failure_types: Optional[Union[FailureType, List[FailureType]]] = None,
        tools: Optional[Union[str, List[str]]] = None,
    ):
        self.url = url
        super().__init__(path / "phm_data_challenge_2018", "RUL")
        self._prepare_dataset()
        self.failure_types = failure_types
        self.tools = tools

        if self.failure_types is not None:
            if not isinstance(self.failure_types, list):
                self.failure_types = [failure_types]
            self.cycles_metadata = self.cycles_metadata[
                self.cycles_metadata["Fault name"].isin(
                    [f.value for f in self.failure_types]
                )
            ]
 
        
        if self.tools is not None:
            if not isinstance(self.tools, list):
                self.tools = [tools]
            self.cycles_metadata = self.cycles_metadata[
                self.cycles_metadata["Tool"].isin(self.tools)
            ]

    def _prepare_dataset(self):
        if self.cycles_table_filename.is_file():
            return
        if not (self.dataset_path / "raw" / "train").is_dir():
            self.prepare_raw_dataset()
        files = list(Path(self.dataset_path / "raw" / "train").resolve().glob("*.csv"))
        faults_files = list(
            Path(self.dataset_path / "raw" / "train" / "train_faults")
            .resolve()
            .glob("*.csv")
        )

        def get_key_from_filename(filename: str) -> str:
            return "_".join(filename.split("_")[0:2])

        fault_files_map = {get_key_from_filename(f.name): f for f in faults_files}
        data_fault_pairs = [
            (file, fault_files_map[get_key_from_filename(file.name)]) for file in files
        ]

        (
            DatasetBuilder()
            .set_splitting_method(
                FailureDataCycleSplitter(
                    data_time_column="time", fault_time_column="time"
                )
            )
            .set_rul_column_method(NumberOfRowsRULColumn())
            .set_output_mode(
                LocalStorageOutputMode(
                    self.dataset_path, output_format=DatasetFormat.PARQUET
                ).set_metadata_columns(
                    {"Tool": "Tool_data", "Fault name": "fault_name"}
                )
            )
            .set_index_column("time")
            .prepare_from_data_fault_pairs_files(
                data_fault_pairs,
            )
        )

    def prepare_raw_dataset(self):
        """Download and unzip the raw files

        Args:
            path (Path): Path where to store the raw dataset
        """

        def track_progress(members):
            for member in tqdm(members, total=70):
                yield member

        path = self.dataset_path / "raw"
        path.mkdir(parents=True, exist_ok=True)
        if not (path / OUTPUT).resolve().is_file():
            download(self.url, path)
        logger.info("Decompressing  dataset...")
        with tarfile.open(path / OUTPUT, "r") as tarball:

            def is_within_directory(directory, target):
                abs_directory = os.path.abspath(directory)
                abs_target = os.path.abspath(target)
                prefix = os.path.commonprefix([abs_directory, abs_target])
                return prefix == abs_directory

            def safe_extract(tar, path=".", members=None, *, numeric_owner=False):
                for member in tar.getmembers():
                    member_path = os.path.join(path, member.name)
                    if not is_within_directory(path, member_path):
                        raise Exception("Attempted Path Traversal in Tar File")

                tar.extractall(path, members, numeric_owner=numeric_owner)

            safe_extract(tarball, path=path, members=track_progress(tarball))
        shutil.move(
            str(path / "phm_data_challenge_2018" / "train"), str(path / "train")
        )
        shutil.move(str(path / "phm_data_challenge_2018" / "test"), str(path / "test"))
        shutil.rmtree(str(path / "phm_data_challenge_2018"))
        (path / OUTPUT).unlink()



from copy import deepcopy
from typing import Any, List, Optional, Tuple, Union

import pandas as pd
from ceruleo.transformation.functional.graph_utils import (
    root_nodes,
    topological_sort_iterator,
)
from ceruleo.transformation.functional.pipeline.pipeline import Pipeline
from ceruleo.transformation.functional.transformerstep import TransformerStep


class Joiner(TransformerStep):
    """
    Join multiple run-to-failure cycles into a single DataFrame
    """
    def transform(self, X: List[pd.DataFrame]) -> pd.DataFrame:
        """
        Join the input lifes

        Parameters:
            X: List of run-to-failure cycles to join

        Returns:
            A dataframe with the joined run-to-failure cycles
        """
        if isinstance(X, list):
            X_default = X[0]
            X_q = pd.concat(X[1:])
            missing_indices = X_default.index.difference(X_q.index)
            X_q = pd.concat((X_q, X_default.loc[missing_indices, :])).sort_index()
            return X_q
        else:
            return X

class Filter(TransformerStep):
    """
    Filter rows of a dataframe based on a query

    Parameters:
        values: Values to filter by
        columns: Columns to filter by
    """
    def __init__(
        self,
        *,
        values: List[Any],
        columns: Union[List[str], str],
        name: Optional[str] = None,
    ):
        def prepare_value(v):
            if isinstance(v, str):
                return f"'{v}'"
            else:
                return v

        super().__init__(name=name)
        self.values = values
        self.columns = columns
        self.query = " & ".join(
            [f"({c} == {prepare_value(v)})" for c, v in zip(self.columns, self.values)]
        )

    def transform(self, X:pd.DataFrame) -> pd.DataFrame:
        """ 
        Filter the dataframe

        Parameters:
            X: Input dataframe
        
        Returns:
            A dataframe with the filtered rows
        """
        if self.values == ["__category_all__"]:
            return X.drop(columns=self.columns)
        else:
            return X.query(self.query).drop(columns=self.columns)


class SplitByCategory(TransformerStep):
    def __init__(
        self,
        *,
        features: Union[str, List[str]],
        pipeline: Pipeline,
        add_default: bool = True,
        name: Optional[str] = None,
    ):
        super().__init__(name=name)
        if isinstance(features, str):
            features = [features]
        self.add_default = add_default
        self.orig_pipeline = deepcopy(pipeline)
        self.categorical_feature_names = features

        self._categorical_feature_names_resolved = None
        self._sub_pipelines = {}
        self.joiner = Joiner()(self)

    def _build_pipeline(self, categories: Tuple[str]):
        self.disconnect(self.joiner)
        s = Filter(values=categories, columns=self._categorical_feature_names_resolved)(
            self
        )
        new_pipe = deepcopy(self.orig_pipeline)
        for node in topological_sort_iterator(new_pipe):
            node.name = f"Category: {categories} " + node.name
        for r in root_nodes(new_pipe):
            r(s)
        self.joiner(new_pipe.final_step)
        return new_pipe

    def transform(self, X):
        return X

    def partial_fit(self, X, y=None):
        if self._categorical_feature_names_resolved is None:
            self._categorical_feature_names_resolved = [
                self.find_feature(X, f) for f in self.categorical_feature_names
            ]
        if self.add_default and "default" not in self._sub_pipelines:
            self._sub_pipelines["default"] = self._build_pipeline(["__category_all__"])
        for _, c in (
            X[self._categorical_feature_names_resolved].drop_duplicates().iterrows()
        ):
            c = tuple(c)
            if c not in self._sub_pipelines:
                self._sub_pipelines[c] = self._build_pipeline(c)
        return self

    def __call__(
        self, prev: Union[List["TransformerStepMixin"], "TransformerStepMixin"]
    ):
        self._add_previous(prev)
        return self.joiner



"""Batching capabilities

Many machine learning models require data to be provided in the form of mini batches.
This module interacts with iterators to generate batches.
In particular, this module was created to be used with pytorch models.

Take a look at the pytorch example to see its usage.
"""
import math
from typing import Tuple

import numpy as np
import pandas as pd
from numpy.lib.arraysetops import isin
from ceruleo.dataset.ts_dataset import AbstractPDMDataset
from ceruleo.iterators.iterators import (
    NotWeighted,
    SampleWeight,
    WindowedDatasetIterator,
)
from ceruleo.iterators.shufflers import AbstractShuffler, NotShuffled


class Batcher:
    """
    WindowedIterator Batcher   

    Parameters:
        iterator: Dataset iterator
        batch_size: int

    """

    def __init__(
        self,
        iterator: WindowedDatasetIterator,
        batch_size: int,
    ):
        self.iterator = iterator
        self.batch_size = batch_size
        self.stop = False
        self.batch_data = None

    @staticmethod
    def new(
        dataset: AbstractPDMDataset,
        window: int,
        batch_size: int,
        step: int,
        horizon: int = 1,
        shuffler: AbstractShuffler = NotShuffled(),
        sample_weight: SampleWeight = NotWeighted(),
        right_closed: bool = True,
        padding: bool = False,
    ) -> "Batcher":
        """
        Batcher constructor from a dataset

        The method constructs a WindowedDatasetIterator from the dataset and
        then a Batcher from the iterator.
        Most of the parameters come from the WindowedDatasetIterator,


        Example:
            ```
            batcher = Batcher.new(transformed_dataset,
                                window=150,
                                batch_size=64,
                                step=1,
                                horizon=1)
            X, y, data = next(batcher)     
            X.shape

            (64, 150, n_features)       
            ```                         

        Parameters:
            dataset: Dataset from which the batcher will be created
            batch_size: Batch size
            step: strides
            horizon: Size of the horizon to predict.
            shuffle: AbstractShuffler
            sample_weight: SampleWeight
            right_closed: bool
            padding: wheter to pad data if there are not enough points to fill the window

        Returns:
            A new constructed batcher
        """
        iterator = WindowedDatasetIterator(
            dataset,
            window,
            step=step,
            horizon=horizon,
            shuffler=shuffler,
            sample_weight=sample_weight,
            right_closed=right_closed,
            padding=padding,
        )
        b = Batcher(iterator, batch_size)
        return b

    def __len__(self) -> int:
        """
        Number of batches

        Returns:
            Number of batches in the iterator
        """
        if len(self.iterator) is None:
            return None
        q = math.ceil(len(self.iterator) / self.batch_size)
        return q

    def __iter__(self):
        self.stop = False
        self.iterator.__iter__()
        return self

    @property
    def n_features(self) -> int:
        """
        Number of features of the transformed dataset

        This is a helper method to obtain the transformed dataset information from the WindowedDatasetIterator

        Returns:
           Number of features of the transformed dataset
        """
        return self.iterator.n_features

    @property
    def window_size(self) -> int:
        """
        Lookback window size

        This is a helper method to obtain the WindowedDatasetIterator information

        Returns:
            Lookback window size
        """
        return self.iterator.window_size

    @property
    def output_shape(self) -> int:
        """
        Number of values returned as target by each sample

        Returns:
            Number of values returned as target by each sample
        """
        return self.iterator.output_size

    @property
    def input_shape(self) -> Tuple[int, int]:
        """
        Tuple containing (window_size, n_features)

        Returns:
            (window_size, n_features)
        """
        return self.iterator.input_shape

    @property
    def computed_step(self):
        if isinstance(self.step, int):
            return self.step
        elif isinstance(self.step, tuple):
            if self.step[0] == "auto":
                return int(self.window / self.step[1])
        raise ValueError("Invalid step parameter")

    def initialize_batch(self):
        def initialize_batch_element(elem):
            if isinstance(elem, tuple):
                for e in elem:
                    initialize_batch_element(e)
            else:
                elem.fill(0)

        if self.batch_data is None:
            return
        for i in range(len(self.batch_data)):
            initialize_batch_element(self.batch_data[i])

    def allocate_batch_data(self, d):
        def allocate_batch_data_element(d):
            if isinstance(d, tuple):
                return tuple(allocate_batch_data_element(q) for q in d)
            else:

                if isinstance(d, np.ndarray) or isinstance(d, pd.Series):
                    shape = d.shape
                elif isinstance(d, list):
                    shape = (len(d),)
                return np.zeros((self.batch_size, *shape))

        if self.batch_data is not None:
            return
        self.batch_data = []
        for i in range(len(d)):
            self.batch_data.append(allocate_batch_data_element(d[i]))

    def _assign_data(self, d, j):
        for i, elem in enumerate(d):
            if isinstance(elem, tuple):
                for k in range(len(elem)):
                    self.batch_data[i][k][j, :] = elem[k]
            else:
                self.batch_data[i][j, :] = elem

    def _slice_data(self, actual_batch_size):
        def slice_batch_data_element(d, actual_batch_size):
            if isinstance(d, tuple):
                return tuple(slice_batch_data_element(q, actual_batch_size) for q in d)
            else:
                return d[: actual_batch_size - 1, :]

        if actual_batch_size == self.batch_size:
            return self.batch_data
        sliced_data = []
        for i in range(len(self.batch_data)):
            sliced_data.append(
                slice_batch_data_element(self.batch_data[i], actual_batch_size)
            )
        return sliced_data

    def __next__(self):
        if self.stop:
            raise StopIteration
        try:
            actual_batch_size = 0
            for j in range(self.batch_size):
                actual_batch_size += 1
                d = next(self.iterator)
                self.allocate_batch_data(d)
                self._assign_data(d, j)
        except StopIteration:
            self.stop = True

        return self._slice_data(actual_batch_size)



from typing import List

import numpy as np
import pandas as pd
import pytest
import scipy.stats

from ceruleo.dataset.ts_dataset import AbstractPDMDataset
from ceruleo.iterators.iterators import RelativeToEnd, RelativeToStart
from ceruleo.transformation.features.entropy import LocalEntropyMeasures
from ceruleo.transformation.features.extraction import (
    EMD,
    ChangesDetector,
    Difference,
    ExpandingStatistics,
    OneHotCategorical,
    RollingStatistics,
    SimpleEncodingCategorical,
    SlidingNonOverlappingEMD,
)
from ceruleo.transformation.features.outliers import (
    EWMAOutOfRange,
    IQROutlierRemover,
    IsolationForestOutlierRemover,
    ZScoreOutlierRemover,
)
from ceruleo.transformation.features.resamplers import IntegerIndexResamplerTransformer
from ceruleo.transformation.features.selection import (
    ByNameFeatureSelector,
    NullProportionSelector,
)
from ceruleo.transformation.features.slicing import SliceRows
from ceruleo.transformation.features.transformation import (
    Accumulate,
    Apply,
    Clip,
    Diff,
    ExpandingCentering,
    ExpandingNormalization,
    MeanCentering,
    MedianCentering,
    Peaks,
    RollingCentering,
    Scale,
    Sqrt,
    Square,
    StringConcatenate,
    SubstractLinebase,
)
from ceruleo.transformation.functional.pipeline.pipeline import Pipeline
from ceruleo.transformation.utils import (
    IdentityTransformerStep,
    PandasToNumpy,
    QuantileEstimator,
    TransformerLambda,
)


def manual_expanding(df: pd.DataFrame, min_points: int = 1):
    to_compute = [
        "kurtosis",
        "skewness",
        "max",
        "min",
        "std",
        "peak",
        "impulse",
        "clearance",
        "rms",
        "shape",
        "crest",
    ]
    dfs = []
    for c in df.columns:
        d = []
        for i in range(min_points - 1):
            d.append([np.nan for f in to_compute])
        for end in range(min_points, df.shape[0] + 1):
            data = df[c].iloc[:end]
            row = [manual_features(data, f) for f in to_compute]
            d.append(row)
        dfs.append(pd.DataFrame(d, columns=[f"{c}_{f}" for f in to_compute]))
    return pd.concat(dfs, axis=1)


def manual_rolling(df: pd.DataFrame, min_points: int = 1, window_size: int = 5):
    to_compute = [
        "kurtosis",
        "skewness",
        "max",
        "min",
        "std",
        "peak",
        "impulse",
        "clearance",
        "rms",
        "shape",
        "crest",
    ]
    dfs = []
    for c in df.columns:
        d = []
        for i in range(min_points - 1):
            d.append([np.nan for f in to_compute])
        for end in range(min_points, df.shape[0] + 1):
            data = df[c].iloc[max(end - window_size, 0) : end]
            row = [manual_features(data, f) for f in to_compute]
            d.append(row)
        dfs.append(pd.DataFrame(d, columns=[f"{c}_{f}" for f in to_compute]))
    return pd.concat(dfs, axis=1)


def kurtosis(s: pd.Series) -> float:
    return scipy.stats.kurtosis(s.values, bias=False)


def skewness(s: pd.Series) -> float:
    return scipy.stats.skew(s.values, bias=False)


def _max(s: pd.Series) -> float:
    return np.max(s.values)


def _min(s: pd.Series) -> float:
    return np.min(s.values)


def std(s: pd.Series) -> float:
    return np.std(s.values, ddof=1)


def peak(s: pd.Series) -> float:
    return max(s) - min(s)


def impulse(s: pd.Series) -> float:
    return peak(s) / np.mean(np.abs(s))


def clearance(s: pd.Series) -> float:
    return peak(s) / (np.mean(np.sqrt(np.abs(s))) ** 2)


def rms(s: pd.Series) -> float:
    return np.sqrt(np.mean(s**2))


def shape(s: pd.Series) -> float:
    return rms(s) / np.mean(np.abs(s))


def crest(s: pd.Series) -> float:
    return peak(s) / rms(s)


feature_functions = {
    "crest": crest,
    "shape": shape,
    "rms": rms,
    "clearance": clearance,
    "impulse": impulse,
    "peak": peak,
    "std": std,
    "min": _min,
    "max": _max,
    "skewness": skewness,
    "kurtosis": kurtosis,
}


def manual_features(s: pd.Series, name: str):
    return feature_functions[name](s)


class DatasetFromPandas(AbstractPDMDataset):
    def __init__(self, lives: List[pd.DataFrame]):

        self.lives = lives

    def get_time_series(self, i: int):
        return self.lives[i]

    @property
    def rul_column(self):
        return "RUL"

    @property
    def n_time_series(self):
        return len(self.lives)


class MockDataset(AbstractPDMDataset):
    def __init__(self, nlives: int):

        self.lives = [
            pd.DataFrame(
                {
                    "feature1": np.linspace(0, 100, 50),
                    "feature2": np.random.randint(2, size=(50,)),
                    "RUL": np.linspace(100, 0, 50),
                }
            )
            for i in range(nlives)
        ]

    def get_time_series(self, i: int):
        return self.lives[i]

    @property
    def rul_column(self):
        return "RUL"

    @property
    def n_time_series(self):
        return len(self.lives)


class MockDataset2(AbstractPDMDataset):
    def __init__(self, nlives: int):
        N = 500
        self.lives = [
            pd.DataFrame(
                {
                    "feature1": np.random.randn(N) * 0.5 + 2,
                    "feature2": np.random.randn(N) * 0.5 + 5,
                    "RUL": np.linspace(100, 0, N),
                }
            )
            for i in range(nlives - 1)
        ]

        self.lives.append(
            pd.DataFrame(
                {
                    "feature1": np.random.randn(N) * 0.5 + 1,
                    "feature2": np.random.randn(N) * 0.5 + 5,
                    "feature3": np.random.randn(N) * 0.5 + 2,
                    "feature4": np.random.randn(N) * 0.5 + 4,
                    "RUL": np.linspace(100, 0, N),
                }
            )
        )

        for j in range(self.nlives - 1):
            for k in range(5):
                p = np.random.randint(N)
                self.lives[j]["feature1"][p] = 5000
                self.lives[j]["feature2"][p] = 5000

        for k in range(5):
            p = np.random.randint(N)
            self.lives[-1]["feature1"][p] = 5000
            self.lives[-1]["feature2"][p] = 5000
            self.lives[-1]["feature3"][p] = 5000
            self.lives[-1]["feature4"][p] = 5000

    def get_life(self, i: int):
        return self.lives[i]

    @property
    def rul_column(self):
        return "RUL"

    @property
    def nlives(self):
        return len(self.lives)


class TestTransformers:
    def test_IQROutlierRemover(self):

        remover = IQROutlierRemover(clip=False)
        df = pd.DataFrame(
            {
                "a": [0, 0.5, 0.2, 0.1, 0.9, 15, 0.5, 0.3, 0.5],
                "b": [5, 6, 7, 5, 9, 5, 6, 5, -45],
            }
        )
        df_new = remover.fit_transform(df)
        assert np.isposinf(df_new["a"][5])
        assert np.isneginf(df_new["b"][8])

        remover = IQROutlierRemover(clip=False, prefer_partial_fit=True)
        df = pd.DataFrame(
            {
                "a": [0, 0.5, 0.2, 0.1, 0.9, 15, 0.5, 0.3, 0.5],
                "b": [5, 6, 7, 5, 9, 5, 6, 5, -45],
            }
        )
        df_new = remover.fit_transform(df)
        assert np.isposinf(df_new["a"][5])
        assert np.isneginf(df_new["b"][8])

    def test_ZScoreOutlierRemover(self):

        remover = ZScoreOutlierRemover(number_of_std_allowed=2)
        df = pd.DataFrame(
            {
                "a": [0, 0.5, 0.2, 0.1, 0.9, 15, 0.5, 0.3, 0.5],
                "b": [5, 6, 7, 5, 9, 5, 6, 5, 45],
            }
        )
        df_new = remover.fit_transform(df)
        assert pd.isnull(df_new["a"][5])
        assert pd.isnull(df_new["b"][8])

        remover = ZScoreOutlierRemover(number_of_std_allowed=2, prefer_partial_fit=True)
        df = pd.DataFrame(
            {
                "a": [0, 0.5, 0.2, 0.1, 0.9, 15, 0.5, 0.3, 0.5],
                "b": [5, 6, 7, 5, 9, 5, 6, 5, 45],
            }
        )
        df_new = remover.fit_transform(df)
        assert pd.isnull(df_new["a"][5])
        assert pd.isnull(df_new["b"][8])


class TestResamplers:
    def test_resampler(self):

        remover = SubSampleTransformer(2)
        df = pd.DataFrame(
            {
                "a": [0, 0.5, 0.2, 0.1, 0.9, 15, 0.5, 0.3, 0.5, 12],
                "b": [5, 6, 7, 5, 9, 5, 6, 5, 45, 12],
            }
        )
        df_new = remover.fit_transform(df)

        assert df_new.shape[0] == int(df.shape[0] / 2)


class TestSelection:
    def test_selection(self):
        df = pd.DataFrame(
            {
                "a": [0, 0.5, None, None],  # 0.5
                "b": [5, None, None, None],  # 0.75
                "c": [5, 2, 3, None],  # 0.25
            }
        )

        selector = NullProportionSelector(max_null_proportion=0.3)
        df_new = selector.fit_transform(df)
        assert set(df_new.columns) == set(["c"])

        selector = NullProportionSelector(max_null_proportion=0.55)
        df_new = selector.fit_transform(df)
        assert set(df_new.columns) == set(["a", "c"])

        selector = NullProportionSelector(max_null_proportion=0.8)
        df_new = selector.fit_transform(df)
        assert set(df_new.columns) == set(["a", "b", "c"])


class TestGenerators:
    def test_EMD(self):
        emd = EMD(n=3)
        df = pd.DataFrame(
            {
                "feature1": np.linspace(-50, 50, 500),
                "feature2": np.cos(np.linspace(-1, 50, 500)),
            }
        )
        df_t = emd.fit_transform(df)
        assert len(df_t["feature1_1"].dropna()) == 0
        assert df_t.shape[1] == 6

    def test_ChangesCounter(self):

        df = pd.DataFrame(
            {
                "feature1": ["a", "a", "a", "b", "b", "a", "a", "c"],
                "feature2": ["a", "a", "b", "b", "c", "c", "c", "c"],
            }
        )
        df_gt = pd.DataFrame(
            {"feature1": [1, 0, 0, 1, 0, 1, 0, 1], "feature2": [1, 0, 1, 0, 1, 0, 0, 0]}
        )
        t = ChangesDetector()
        df1 = t.fit_transform(df).astype("int")
        assert df1.compare(df_gt).empty

    def test_Accumulate(self):
        df = pd.DataFrame(
            {"a": [1, 2, 3, 4], "b": [2, 4, 6, 8], "c": [2, 2, 2, 2], "d": [1, 0, 1, 0]}
        )
        transformer = Accumulate()
        df_new = transformer.fit_transform(df)

        assert df_new["a"].iloc[-1] == 10
        assert df_new["b"].iloc[-1] == 20
        assert df_new["c"].iloc[-1] == 8
        assert (df_new["d"].values == np.array([1, 1, 2, 2])).all()

        ds = MockDataset(5)
        transformer = Accumulate()
        transformer.fit(ds)

        life_0 = transformer.transform(ds.lives[0])
        life_1 = transformer.transform(ds.lives[1])

        assert (life_0["feature1"] == ds.lives[0]["feature1"].cumsum()).all()
        assert (life_0["feature2"] == ds.lives[0]["feature2"].cumsum()).all()

        assert (life_1["feature1"] == ds.lives[1]["feature1"].cumsum()).all()
        assert (life_1["feature2"] == ds.lives[1]["feature2"].cumsum()).all()

    def test_expanding(self):

        lives = [
            pd.DataFrame(
                {
                    "a": np.random.rand(50) * 100 * np.random.rand(50) ** 3,
                    "b": np.random.rand(50) * 100 * np.random.rand(50) ** 2,
                }
            )
        ]
        to_compute = [
            "kurtosis",
            "skewness",
            "max",
            "min",
            "std",
            "peak",
            "impulse",
            "clearance",
            "rms",
            "shape",
            "crest",
        ]
        expanding = ExpandingStatistics(to_compute=to_compute)

        ds_train = DatasetFromPandas(lives[0:2])
        ds_test = DatasetFromPandas(lives[2:2])

        expanding.fit(ds_train)

        pandas_t = expanding.transform(ds_train[0][["a", "b"]])
        fixed_t = manual_expanding(ds_train[0][["a", "b"]], 2)

        assert (pandas_t - fixed_t).mean().mean() < 1e-15

    def test_rolling(self):
        lives = [
            pd.DataFrame(
                {
                    "a": np.random.rand(50) * 100 * np.random.rand(50) ** 3,
                    "b": np.random.rand(50) * 100 * np.random.rand(50) ** 2,
                }
            )
        ]
        to_compute = [
            "kurtosis",
            "skewness",
            "max",
            "min",
            "std",
            "peak",
            "impulse",
            "clearance",
            "rms",
            "shape",
            "crest",
        ]
        rolling = RollingStatistics(to_compute=to_compute, min_points=2, window=5)

        ds_train = DatasetFromPandas(lives[0:2])
        ds_test = DatasetFromPandas(lives[2:2])

        rolling.fit(ds_train)

        pandas_t = rolling.transform(ds_train[0][["a", "b"]])
        fixed_t = manual_rolling(ds_train[0][["a", "b"]], 2, 5)

        assert (pandas_t - fixed_t).mean().mean() < 1e-10

        rolling = RollingStatistics(
            min_points=2,
            window=5,
            specific={"a": ["mean", "kurtosis"], "b": ["peak", "impulse"]},
        )
        rolling.fit(ds_train)
        pandas_t = rolling.transform(ds_train[0][["a", "b"]])
        assert sorted(pandas_t.columns) == sorted(['a_mean', 'a_kurtosis', 'b_peak', 'b_impulse'])

    def test_EWMAOutOfRange(self):
        a = np.random.randn(500) * 0.5 + 2
        b = np.random.randn(500) * 0.5 + 5
        a[120] = 1500
        a[320] = 5000

        b[120] = 1500
        b[320] = 5000
        b[215] = 1500

        df = pd.DataFrame(
            {
                "a": a,
                "b": b,
            }
        )
        transformer = ByNameFeatureSelector(features=["a", "b"])
        transformer = EWMAOutOfRange(return_mask=True)(transformer)
        transformer = Accumulate()(transformer)
        df_new = Pipeline(transformer).fit_transform(df)
        assert df_new["a"].iloc[-1] == 2
        assert df_new["b"].iloc[-1] == 3

        transformer = ByNameFeatureSelector(features=["a", "b"])
        transformer = EWMAOutOfRange(return_mask=False)(transformer)
        df_new = Pipeline(transformer).fit_transform(df)

        assert np.isnan(df_new["a"].iloc[320])
        assert np.isnan(df_new["b"].iloc[215])

    def test_IsolationForestOutlierRemover(self):
        a = np.random.randn(500) * 0.5 + 2
        b = np.random.randn(500) * 0.5 + 5
        a[120] = 1500
        a[320] = 5000

        b[120] = 1500
        b[320] = 5000
        b[215] = 1500

        df = pd.DataFrame(
            {
                "a": a,
                "b": b,
            }
        )
        transformer = ByNameFeatureSelector(features=["a", "b"])
        transformer = IsolationForestOutlierRemover()(transformer)
        df_new = Pipeline(transformer).fit_transform(df)

        assert np.isnan(df_new["a"].iloc[320])
        assert np.isnan(df_new["b"].iloc[215])

    def test_encodings(self):
        df = pd.DataFrame(
            {
                "a": ["c1", "c2", "c3", "c1", "c3"],
                "b": [1, 1, 1, 1, 1],
            }
        )
        df1 = pd.DataFrame(
            {
                "a": ["c1", "c1", "c2", "c1", "c4"],
                "b": [1, 1, 1, 1, 1],
            }
        )
        df2 = pd.DataFrame(
            {
                "a": ["c1", "c1", "c1", "c3", "c2"],
                "b": [1, 1, 1, 1, 1],
            }
        )
        df3 = pd.DataFrame(
            {
                "a": ["c1", "c1", "c1", "c3", "c5"],
                "b": [1, 1, 1, 1, 1],
            }
        )
        transformer = OneHotCategorical(feature="a")
        transformer.partial_fit(df)
        transformer.partial_fit(df1)

        df_t = transformer.transform(df2)
        df_true = pd.DataFrame(
            {
                "c1": [1, 1, 1, 0, 0],
                "c2": [0, 0, 0, 0, 1],
                "c3": [0, 0, 0, 1, 0],
                "c4": [0, 0, 0, 0, 0],
            }
        )

        assert (df_t == df_true).all().all()

        df_t = transformer.transform(df3)
        df_true = pd.DataFrame(
            {
                "c1": [1, 1, 1, 0, 0],
                "c2": [0, 0, 0, 0, 0],
                "c3": [0, 0, 0, 1, 0],
                "c4": [0, 0, 0, 0, 0],
            }
        )

        assert (df_t == df_true).all().all()

        transformer = SimpleEncodingCategorical(feature="a")
        transformer.partial_fit(df)
        transformer.partial_fit(df1)

        df_t = transformer.transform(df2)
        df_true = pd.DataFrame([0, 0, 0, 2, 1])

        assert np.all(df_true.values == df_t.values)

        df_t = transformer.transform(df3)
        df_true = pd.DataFrame([0, 0, 0, 2, -1])
        assert np.all(df_true.values == df_t.values)

    def test_Difference(self):

        df = pd.DataFrame(
            {"a": [1, 2, 3, 4], "b": [2, 4, 6, 8], "c": [2, 2, 2, 2], "d": [1, 1, 1, 1]}
        )
        with pytest.raises(ValueError):
            transformer = Difference(feature_set1=["a", "b"], feature_set2=["d"])

        transformer = Difference(feature_set1=["a", "b"], feature_set2=["c", "d"])
        df_new = transformer.fit_transform(df)
        assert (df_new["a"].values == np.array([1 - 2, 2 - 2, 3 - 2, 4 - 2])).all()
        assert (df_new["b"].values == np.array([2 - 1, 4 - 1, 6 - 1, 8 - 1])).all()
        assert df_new.shape[1] == 2


class TestEntropy:
    def test_entropy(self):
        df = pd.DataFrame({"a": [0, 0, 1, 1, 1, 1, 0, 0, 0]})

        selector = LocalEntropyMeasures(2)
        df_new = selector.fit_transform(df)

        LOCAL_BLOCK_ENTROPY = np.array(
            [
                np.nan,
                1.4150375,
                3.0,
                1.4150375,
                1.4150375,
                1.4150375,
                3.0,
                1.4150375,
                1.4150375,
            ]
        )
        LOCAL_ENTROPY_RATE = np.array(
            [np.nan, np.nan, 1, 0.0, 0.5849625, 0.5849625, 1.5849625, 0.0, 1.0]
        )
        LOCAL_ACTIVE_INFORMATION = np.array(
            [
                np.nan,
                np.nan,
                -0.19264508,
                0.80735492,
                0.22239242,
                0.22239242,
                -0.36257008,
                1.22239242,
                0.22239242,
            ]
        )
        assert np.nansum(df_new["a_local_entropy_rate"] - LOCAL_ENTROPY_RATE) < 1e-6

        assert np.nansum(df_new["a_local_block_entropy"] - LOCAL_BLOCK_ENTROPY) < 1e-6

        assert (
            np.nansum(df_new["a_local_active_information"] - LOCAL_ACTIVE_INFORMATION)
            < 1e-6
        )


class TestQuantileEstimator:
    def test_quantile(self):

        A = pd.DataFrame({"A": np.random.randn(15000)*10, "B": np.random.randn(15000)*10})
        q = QuantileEstimator(tdigest_size=200)
        q.update(A)
        s = q.estimate_quantile(0.5)
        assert s.index.tolist() == ["A", "B"]
        assert (np.abs(s.A - A.quantile(0.5)["A"])) < 0.1
        assert (np.abs(s.B - A.quantile(0.5)["B"])) < 0.1

        s = q.estimate_quantile(0.1)

        assert (np.abs(s.A - A.quantile(0.1)["A"])) < 0.1
        assert (np.abs(s.B - A.quantile(0.1)["B"])) < 0.1

        B = pd.DataFrame({"A": np.random.randn(15000) + 5, "B": np.random.randn(15000)})
        q.update(B)
        s = q.estimate_quantile(0.5)
        assert s.index.tolist() == ["A", "B"]
        assert np.abs(s.A) - 5 < 0.1
        assert np.abs(s.B) < 0.1
        assert s.index.tolist() == ["A", "B"]

        assert q.estimate_quantile(0.5, "A") - 5 < 0.1


class TestEMD:
    def test_emd(self):
        A = pd.DataFrame({"A": np.random.randn(15000), "B": np.random.randn(15000)})
        q = SlidingNonOverlappingEMD(window_size=300, max_imfs=5)
        r = q.transform(A)
        assert r.shape[0] == A.shape[0]
        assert r.shape[1] == A.shape[1] * 5


class TestResamplers:
    def test_resampler(self):

        A = pd.DataFrame(
            {
                "time": np.linspace(500, 0, 25).astype(int),
                "B": np.linspace(0, 500, 25).astype(int),
            }
        )
        A.loc[5, "time"] = A.loc[4, "time"] - 5
        A.loc[10, "time"] = A.loc[9, "time"] - 12
        A.loc[17, "time"] = A.loc[16, "time"] - 35

        resampler = IntegerIndexResamplerTransformer(
            time_feature="time", steps=15, drop_time_feature=True
        )

        resampler.partial_fit(A)
        q = resampler.transform(A)
        assert q.columns.values.tolist() == ["B"]
        assert np.all(np.diff(q) == 15)

        resampler = IntegerIndexResamplerTransformer(
            time_feature="time", steps=15, drop_time_feature=False
        )

        resampler.partial_fit(A)
        q = resampler.transform(A)
        assert q.columns.values.tolist() == ["time", "B"]

        A = pd.DataFrame({"time": [5, 3, 0], "B": [10, 6, 0]})
        resampler = IntegerIndexResamplerTransformer(
            time_feature="time", steps=2, drop_time_feature=False
        )

        resampler.partial_fit(A)
        q = resampler.transform(A)
        assert np.sum(q["B"] - np.array([0, 4, 8])) < 0.000005


def test_utils():

    input = pd.DataFrame({"a": [1, 2, 3]})
    out = PandasToNumpy().fit_transform(input)
    assert isinstance(out, np.ndarray)
    assert out.shape == (3, 1)

    out = TransformerLambda(callback=lambda x: x*2).fit_transform(input)
    assert out["a"].tolist() == [2, 4, 6]

    out = IdentityTransformerStep().fit_transform(input)
    assert out["a"].tolist() == [1, 2, 3]

    out = IdentityTransformerStep().fit_transform(input.values)
    assert out.tolist() == [[1], [2], [3]]


def test_transformation():
    input = pd.DataFrame({"a": [5, 10, 15]})
    out = MeanCentering().fit_transform(input)
    assert out["a"].tolist() == [-5, 0, 5]

    out = MedianCentering().fit_transform(input)
    assert out["a"].tolist() == [-5, 0, 5]

    input1 = pd.DataFrame({"a": np.random.randn(200)*5 + 50})
    input2 = pd.DataFrame({"a": np.random.randn(200)*15 + 50})

    
    centering = (MedianCentering()
            .partial_fit(input1)
            .partial_fit(input2)
    )
    
    assert np.abs(centering.median["a"] - 50 ) < 1
    out = centering.transform(input1)
    assert out.equals(input1 - centering.median)

    centering = (MedianCentering()
            .partial_fit(pd.DataFrame({"a": np.random.randn(1)*5 + 50}))
    )
    assert centering.median is None

    square = Square()
    out = square.fit_transform(input)
    assert out["a"].tolist() == [25, 100, 225]

    scaler = Scale(scale_factor=2)
    out = scaler.fit_transform(input)
    assert out["a"].tolist() == [10, 20, 30]
    
    
    expanding_centering = ExpandingCentering()
    out = expanding_centering.fit_transform(input1)
    assert out.equals(input1 - input1.expanding().mean())

    input = pd.DataFrame({"a": np.random.randn(200)*5 + 50})
    rolling_centering = RollingCentering(window=5, min_points=1)
    out = rolling_centering.fit_transform(input)
    assert out.equals(input - input.rolling(window=5, min_periods=1).mean())

    input = pd.DataFrame({"a": np.random.randn(200)*5 + 50})
    sqrt = Sqrt()
    out = sqrt.fit_transform(input)
    assert out["a"].tolist() == np.sqrt(input["a"]).tolist()

    expanding_normalization = ExpandingNormalization()
    out = expanding_normalization.fit_transform(input)
    assert out.equals((input - input.expanding().mean()) / (input.expanding().std()))

    accumulate = Accumulate()
    out = accumulate.fit_transform(input)
    assert out.equals(input.cumsum())

    accumulate = Accumulate(normalize=True)
    out = accumulate.fit_transform(input)
    cummulated = input.cumsum()
    assert out.equals(cummulated / cummulated.abs().apply(np.sqrt))

    peaks = Peaks(distance=3)
    input = pd.DataFrame({"a": np.cos(np.arange(-np.pi, 3*np.pi, np.pi/4))})
    out = peaks.fit_transform(input)
    assert out["a"].sum() == 2
    assert out["a"].iloc[4] == 1
    assert out["a"].iloc[12] == 1

    clip = Clip(lower=0, upper=1)
    input = pd.DataFrame({"a": np.linspace(-1, 2, 10)})
    out = clip.fit_transform(input)
    assert out["a"].min() == 0
    assert out["a"].max() == 1

    substract_linebase = SubstractLinebase()
    input = pd.DataFrame({"a": np.linspace(5, 10, 10)})
    out = substract_linebase.fit_transform(input)
    assert out["a"].min() == 0
    assert out["a"].max() == 5

    diff = Diff()
    input = pd.DataFrame({"a": np.linspace(5, 10, 10)})
    out = diff.fit_transform(input)
    assert out.equals(input.diff())

    apply = Apply(fun=lambda x: x**2)
    input = pd.DataFrame({"a": np.linspace(5, 10, 10)})
    out = apply.fit_transform(input)
    assert out.equals(input.apply(lambda x: x**2))

    string_concat = StringConcatenate()
    input = pd.DataFrame({"a": ["a", "b", "c"],
                          "b": ["d", "e", "f"]})
    out = string_concat.fit_transform(input)
    assert out["concatenation"].tolist() == ["a-d", "b-e", "c-f"]
    



def test_slicing():
    slicer = SliceRows(
        initial = RelativeToStart(5),
        final = RelativeToEnd(5),
    )
    input = pd.DataFrame({"a": np.linspace(5, 10, 10)})
    out = slicer.fit_transform(input)
    assert out.equals(input.iloc[5:-5])


from numpy.lib.arraysetops import isin
from ceruleo.dataset.transformed import TransformedDataset
from ceruleo.dataset.ts_dataset import AbstractPDMDataset
from typing import Optional, Union

import numpy as np
from ceruleo.dataset.utils import iterate_over_target
from ceruleo.iterators.batcher import Batcher
from ceruleo.iterators.iterators import WindowedDatasetIterator

try:
    import tensorflow as tf

    TENSORFLOW = True
except:
    TENSORFLOW = False


def true_values(
    dataset: Union[WindowedDatasetIterator, Batcher, AbstractPDMDataset],
    target_column: Optional[str] = None
) -> np.array:
    """Obtain the true RUL of the dataset after the transformation

    Parameters:

        dataset:  Iterator of the dataset

    Returns:

        true_RUL: target values after the transformation
    """
    from ceruleo.transformation.functional.transformers import TransformerIdentity

    if isinstance(dataset, Batcher):
        dataset = dataset.iterator

    elif isinstance(dataset, AbstractPDMDataset) and not isinstance(dataset, TransformedDataset):
        if target_column is None:
            if not hasattr(dataset, "rul_column"):
                raise ValueError("Please provide a target column to access")
            else:
                target_column = dataset.rul_column
        return np.squeeze(np.concatenate([y for y in iterate_over_target(dataset)]))
    else:
        if TENSORFLOW:
            if isinstance(dataset, tf.data.Dataset):
                dataset = dataset.as_numpy_iterator()
        return np.squeeze(np.concatenate([y for _, y, _ in dataset]))
    raise ValueError("Ivalid dataset used")



import gc
import gzip
import logging
import pickle
import zipfile
from pathlib import Path
from typing import Optional, Union

import h5py
import numpy as np
import pandas as pd
from ceruleo import DATA_PATH
from ceruleo.dataset.ts_dataset import AbstractPDMDataset
from tqdm.auto import tqdm

from ceruleo.utils.download import download

logger = logging.getLogger(__name__)

DATASET_PATH = DATA_PATH / "C_MAPSS2"

URL = "https://ti.arc.nasa.gov/c/47/"


def load_file(filename: Union[Path, str], train: bool = True):
    with h5py.File(filename, "r") as hdf:
        if train:
            W = np.array(hdf.get("W_dev"))  # W
            X_sv = np.array(hdf.get("X_s_dev"))  # X_s
            X_v = np.array(hdf.get("X_v_dev"))  # X_v
            T = np.array(hdf.get("T_dev"))  # T
            Y = np.array(hdf.get("Y_dev"))  # RUL
            A = np.array(hdf.get("A_dev"))  # Auxiliary

        else:
            W = np.array(hdf.get("W_test"))  # W
            X_sv = np.array(hdf.get("X_s_test"))  # X_s
            X_v = np.array(hdf.get("X_v_test"))  # X_v
            T = np.array(hdf.get("T_test"))  # T
            Y = np.array(hdf.get("Y_test"))  # RUL
            A = np.array(hdf.get("A_test"))  # Auxiliary

        # Varnams
        W_var = np.array(hdf.get("W_var"))
        X_s_var = np.array(hdf.get("X_s_var"))
        X_v_var = np.array(hdf.get("X_v_var"))
        T_var = np.array(hdf.get("T_var"))
        A_var = np.array(hdf.get("A_var"))

        # from np.array to list dtype U4/U5
        W_var = list(np.array(W_var, dtype="U20"))
        X_s_var = list(np.array(X_s_var, dtype="U20"))
        X_v_var = list(np.array(X_v_var, dtype="U20"))
        T_var = list(np.array(T_var, dtype="U20"))
        A_var = list(np.array(A_var, dtype="U20"))

        data = pd.DataFrame(
            np.concatenate((W, X_sv, X_v, T, Y, A), axis=1),
            columns=W_var + X_s_var + X_v_var + T_var + ["RUL"] + A_var,
        )
        return data


class CMAPSS2PreProcessor:
    def __init__(self, path: Path = DATASET_PATH):
        self.path = path
        self.raw_data_path = path / "data_set"
        self.lives_table_path = path / "lives_data.pkl"

    def process_raw_dataframe(
        self, data:pd.DataFrame, file_h5: Path, i: int, unit: int, train: bool
    ):
        unit = int(unit)
        output_dir = f"LIFE_{file_h5.name}_{unit}_{i}"
        number_of_samples = data.shape[0]
        data.to_parquet(self.path / output_dir)
        return {
            "Raw File": file_h5.name,
            "Unit": unit,
            "Index": i,
            "Output Dir": output_dir,
            "Train": train,
            "Number of samples": number_of_samples,
        }

    def process_raw_file(self, file_h5: Path, train: bool):
        info = []
        try:
            df = load_file(file_h5, train=train)
            for i, (unit, data) in enumerate(df.groupby(by=["unit"])):
                info.append(self.process_raw_dataframe(data, file_h5, i, unit, train))
            gc.collect()
        except OSError as e:
            logger.error(f"Cannot open file {file_h5}")
            logger.error(e)
        return info

    def process_raw_files(self):
        info = []
        for file_h5 in tqdm(list(self.raw_data_path.glob("*.h5"))):
            info.extend(self.process_raw_file(file_h5, train=True))
            info.extend(self.process_raw_file(file_h5, train=False))

        df = pd.DataFrame(info)
        with open(self.lives_table_path, "wb") as file:
            pickle.dump(df, file)



    def obtain_raw_files(self):
        logger.info("Dataset not processed.")
        if not self.raw_data_path.is_dir():
            ZIP_FILE = self.path / "data_set.zip"
            if not ZIP_FILE.is_file():
                logger.info('Downloading file')
                self.raw_data_path.mkdir(exist_ok=True, parents=True)
                download(URL, ZIP_FILE)
            logger.info("Unzipping")
            with zipfile.ZipFile(ZIP_FILE, "r") as zip_ref:
                zip_ref.extractall(self.path)
            logger.info("Removing zip file")
            ZIP_FILE.unlink()

    def run(self):
        if not (self.lives_table_path).is_file():
            self.obtain_raw_files()
        self.process_raw_files()



class CMAPSS2Dataset(AbstractPDMDataset):
    """C-MAPSS-2 Dataset

    The dataset provides a new realistic dataset of run-to-failure trajectories for a small fleet of aircraft
    engines under realistic flight conditions.

    The damage propagation modelling used for the generation of this synthetic dataset builds on
    the modeling strategy from previous work .
    The dataset was generated with the Commercial Modular Aero-Propulsion System Simulation (C-MAPSS) dynamical model.
    The data set is been provided by the Prognostics CoE at NASA Ames in collaboration with ETH Zurich and PARC.

    [Dataset reference](https://data.phmsociety.org/2021-phm-conference-data-challenge/)

    Parameters:
        train: Wether to obtain the train data provided
    """
    def __init__(
        self,
        path: Path = DATASET_PATH,
        train: Optional[bool] = None,
    ):
        super().__init__()
        self.path = path
        LIVES_TABLE_PATH = path / "lives_data.pkl"
        if not (LIVES_TABLE_PATH).is_file():
            pr = CMAPSS2PreProcessor()
            pr.run()
        with open(LIVES_TABLE_PATH, "rb") as file:
            self.lives = pickle.load(file)

        if train is not None:
            self.lives = self.lives[self.lives["Train"] == train]

    def get_time_series(self, i):
        df_path = self.lives.iloc[i]["Output Dir"]
        df = pd.read_parquet(self.path / df_path)
        return df

    @property
    def n_time_series(self):
        return len(self.lives)

    @property
    def rul_column(self) -> str:
        return "RUL"



from sklearn.ensemble import RandomForestRegressor
from sklearn.linear_model import Ridge
from sklearn.model_selection import GridSearchCV

from ceruleo.dataset.catalog.CMAPSS import CMAPSSDataset, sensor_indices
from ceruleo.iterators.sample_weight import RULInverseWeighted
from ceruleo.models.sklearn import (
    CeruleoMetricWrapper,
    CeruleoRegressor,
    TimeSeriesWindowTransformer,
)
from ceruleo.transformation import Transformer
from ceruleo.transformation.features.scalers import MinMaxScaler
from ceruleo.transformation.features.selection import ByNameFeatureSelector
from ceruleo.transformation.functional.pipeline.pipeline import make_pipeline


def test_gridsearch_cv():
    train_dataset = CMAPSSDataset(train=True, models='FD001')
    FEATURES = [train_dataset[0].columns[i] for i in sensor_indices]
    transformer = Transformer(
        pipelineX=make_pipeline(
            ByNameFeatureSelector(features=FEATURES), 
            MinMaxScaler(range=(-1, 1))

        ), 
        pipelineY=make_pipeline(
            ByNameFeatureSelector(features=['RUL']),  
        )
    )


    regressor_gs = CeruleoRegressor(
        TimeSeriesWindowTransformer(
            transformer,
            window_size=32,
            padding=True,
            step=1),   
        Ridge(alpha=15))

    grid_search = GridSearchCV(
        estimator=regressor_gs,
        param_grid={
            'ts_window_transformer__window_size': [5, 6],         
            'regressor': [Ridge(alpha=15), RandomForestRegressor(max_depth=5)]
        },
        scoring=CeruleoMetricWrapper('neg_mean_absolute_error'),
        cv=2,
        verbose=5
    )

    grid_search.fit(train_dataset)
    assert grid_search is not None



def test_sample_weights():
    train_dataset = CMAPSSDataset(train=True, models='FD001')
    FEATURES = [train_dataset[0].columns[i] for i in sensor_indices]
    transformer = Transformer(
        pipelineX=make_pipeline(
            ByNameFeatureSelector(features=FEATURES), 
            MinMaxScaler(range=(-1, 1))

        ), 
        pipelineY=make_pipeline(
            ByNameFeatureSelector(features=['RUL']),  
        )
    )


    regressor_gs = CeruleoRegressor(
        TimeSeriesWindowTransformer(
            transformer,
            window_size=32,
            sample_weight=RULInverseWeighted(),
            padding=True,
            step=1),   
        Ridge(alpha=15))

   

    regressor_gs = regressor_gs.fit(train_dataset)
    assert regressor_gs is not None



import math

import numpy as np
from typing import Iterable, List


def k_to_q(k: float, d: float) -> float:
    """
    Parameters
    ----------
    k : float
        [description]
    d : float
        [description]

    Returns
    -------
    float
        [description]
    """
    k_div_d = k / d
    if k_div_d >= 0.5:
        base = 1 - k_div_d
        return 1 - 2 * base * base
    else:
        return 2 * k_div_d * k_div_d


def inplace_merge(arr, start, mid, end):
    start2 = mid + 1
    if arr[mid] <= arr[start2]:
        return
    while start <= mid and start2 <= end:
        if arr[start] <= arr[start2]:
            start += 1
        else:
            value = arr[start2]
            index = start2
            while index != start:
                arr[index] = arr[index - 1]
                index -= 1

            arr[start] = value
            start += 1
            mid += 1
            start2 += 1


class Centroid:
    def __init__(self, mean: float = 0.0, weight: float = 1.0):
        assert weight > 0
        self.mean = mean
        self.weight = weight

    def add(self, sum: float, weight: float) -> float:
        """Adds the sum/weight to this centroid, and returns the new sum.

        Parameters
        ----------
        sum : float
            [description]
        weight : float
            [description]

        Returns
        -------
        [type]
            [description]
        """
        sum += self.mean * self.weight
        self.weight += weight
        self.mean = sum / self.weight
        return sum

    def __lt__(self, other):
        return self.mean < other.mean


class TDigest:
    def __init__(self, maxSize: int = 100):
        self.maxSize = maxSize
        self.sum = 0.0
        self.count = 0.0
        self.max = np.nan
        self.min = np.nan
        self.centroids = []

    def construct(
        centroids: List[Centroid],
        sum: float,
        count: float,
        max_val: float,
        min_val: float,
        maxSize: int = 100,
    ):
        t = TDigest(maxSize)
        t.sum = sum
        t.count = count
        t.max = max_val
        t.min = min_val
        if len(centroids) < maxSize:
            t.centroids = centroids
        # // Number of centroids is greater than maxSize, we need to compress them
        # // When merging, resulting digest takes the maxSize of the first digest
        sz = len(centroids)
        digests = [
            TDigest(maxSize),
            TDigest.construct(centroids, sum, count, max, min, sz),
        ]
        return t.merge(digests)

    def estimate_quantile(self, q: float) -> float:
        """Estimates the value of the given quantile.

        Parameters
        ----------
        q : float
            Quantile

        Returns
        -------
        float
            Value of the quantile
        """
        if len(self.centroids) == 0:
            return 0.0
        rank = q * self.count
        pos = 0
        t = 0
        if q > 0.5:
            if q >= 1.0:
                return self.max
            pos = 0
            t = self.count
            for i, centroid in enumerate(self.centroids[::-1]):
                t -= centroid.weight
                if rank >= t:
                    pos = len(self.centroids) - i - 1
                    break

        else:
            if q <= 0.0:
                return self.min

            pos = len(self.centroids) - 1
            t = 0
            for i, centroid in enumerate(self.centroids):
                if rank < t + centroid.weight:
                    pos = i
                    break
                t += centroid.weight

        delta = 0
        min = self.min
        max = self.max
        if len(self.centroids) > 1:
            if pos == 0:
                delta = self.centroids[pos + 1].mean - self.centroids[pos].mean
                max = self.centroids[pos + 1].mean
            elif pos == len(self.centroids) - 1:
                delta = self.centroids[pos].mean - self.centroids[pos - 1].mean
                min = self.centroids[pos - 1].mean
            else:
                delta = (
                    self.centroids[pos + 1].mean - self.centroids[pos - 1].mean
                ) / 2
                min = self.centroids[pos - 1].mean
                max = self.centroids[pos + 1].mean

        value = (
            self.centroids[pos].mean
            + ((rank - t) / self.centroids[pos].weight - 0.5) * delta
        )
        return np.clip(value, min, max)

    def merge(self, digests: List["TDigest"]):
        nCentroids = 0
        for digest in digests:
            nCentroids += digest.centroids_.size()

        if nCentroids == 0:
            return TDigest()

        centroids = []
        starts = []

        count = 0

        # We can safely use these limits to avoid isnan checks below because we know
        # nCentroids > 0, so at least one TDigest has a min and max.
        min = np.inf
        max = -np.inf

        for digest in digests:
            starts.append(len(centroids))
            curCount = digest.count()
            if curCount > 0:
                assert not np.isnan(digest.min)
                assert not np.isnan(digest.max)
                min = min(min, digest.min)
                max = max(max, digest.max)
                count += curCount
                for centroid in digest.centroids:
                    centroids.push_back(centroid)

        digestsPerBlock = 1
        while digestsPerBlock < len(starts):
            # Each sorted block is digestPerBlock digests big. For each step, try to
            # merge two blocks together.
            i = 0
            while i < starts.size():
                # It is possible that this block is incomplete (less than digestsPerBlock
                # big). In that case, the rest of the block is sorted and leave it alone
                if i + digestsPerBlock < starts.size():
                    first = starts[i]
                    middle = starts[i + digestsPerBlock]

                    # It is possible that the next block is incomplete (less than
                    # digestsPerBlock big). In that case, merge to end. Otherwise, merge to
                    # the end of that block.
                    if i + (digestsPerBlock * 2) < len(starts):
                        last = starts[i + 2 * digestsPerBlock]
                    else:
                        last = len(centroids) - 1
                    inplace_merge(centroids, first, middle, last)
            i += digestsPerBlock * 2

            digestsPerBlock *= 2

        maxSize = digests[0].maxSize
        result = TDigest(maxSize)

        compressed = []

        k_limit = 1
        q_limit_times_count = k_to_q(k_limit, maxSize) * count

        cur = centroids[0]
        weightSoFar = cur.weight()
        sumsToMerge = 0
        weightsToMerge = 0
        for centroid in centroids:
            weightSoFar += centroid.weight
            if weightSoFar <= q_limit_times_count:
                sumsToMerge += centroid.mean * centroid.weight
                weightsToMerge += centroid.weight
            else:
                result.sum_ += cur.add(sumsToMerge, weightsToMerge)
                sumsToMerge = 0
                weightsToMerge = 0
                compressed.append(cur)
                q_limit_times_count = k_to_q(k_limit, maxSize) * count
                k_limit += 1
                cur = centroid

        result.sum += cur.add(sumsToMerge, weightsToMerge)
        compressed.append(cur)
        compressed.shrink_to_fit()

        # Deal with floating point precision
        compressed = sorted(compressed)

        result.count = count
        result.min = min
        result.max = max
        result.centroids = compressed
        return result

    def merge_unsorted(self, unsortedValues: Iterable[float]) -> "TDigest":
        """Merge unsorted values by first sorting them.


        Parameters
        ----------
        unsortedValues: Iterable[float]
            Values to Merge

        Returns
        -------
        TDigest
            [description]
        """
        return self.merge_sorted(sorted(unsortedValues))

    def merge_sorted(self, sortedValues: Iterable[float]):
        if len(sortedValues) == 0:
            return self

        result = TDigest(self.maxSize)

        result.count = self.count + len(sortedValues)

        maybeMin = sortedValues[0]
        maybeMax = sortedValues[-1]
        if self.count > 0:
            # // We know that min_ and max_ are numbers
            result.min = min(self.min, maybeMin)
            result.max = max(self.max, maybeMax)
        else:
            # // We know that min_ and max_ are NaN.
            result.min = maybeMin
            result.max = maybeMax

        compressed = []

        k_limit = 1
        q_limit_times_count = k_to_q(k_limit, self.maxSize) * result.count
        k_limit += 1

        it_centroids = 0
        it_sortedValues = 0

        cur = None

        if (it_centroids < len(self.centroids) - 1) and (
            self.centroids[it_centroids].mean < sortedValues[it_sortedValues]
        ):
            cur = self.centroids[it_centroids]
            it_centroids += 1
        else:
            cur = Centroid(sortedValues[it_sortedValues], 1.0)
            it_sortedValues += 1

        weightSoFar = cur.weight

        # Keep track of sums along the way to reduce expensive floating points
        sumsToMerge = 0.0
        weightsToMerge = 0.0

        while (it_centroids < len(self.centroids) - 1) or (
            it_sortedValues < len(sortedValues) - 1
        ):
            next = None

            if (it_centroids < len(self.centroids) - 1) and (
                (it_sortedValues == len(sortedValues) - 1)
                or (self.centroids[it_centroids].mean < sortedValues[it_sortedValues])
            ):
                next = self.centroids[it_centroids]
                it_centroids += 1
            else:
                next = Centroid(sortedValues[it_sortedValues], 1.0)
                it_sortedValues += 1

            nextSum = next.mean * next.weight
            weightSoFar += next.weight

            if weightSoFar <= q_limit_times_count:
                sumsToMerge += nextSum
                weightsToMerge += next.weight
            else:
                result.sum += cur.add(sumsToMerge, weightsToMerge)
                sumsToMerge = 0
                weightsToMerge = 0
                compressed.append(cur)
                q_limit_times_count = k_to_q(k_limit, self.maxSize) * result.count
                k_limit += 1
                cur = next

        result.sum += cur.add(sumsToMerge, weightsToMerge)
        compressed.append(cur)

        # Deal with floating point precision
        compressed = sorted(compressed)

        result.centroids = compressed
        return result



import logging
from typing import Optional

import numpy as np
import pandas as pd
from pydantic import BaseModel
from typing import List 
from ceruleo.dataset.ts_dataset import AbstractPDMDataset
from ceruleo.utils import pydantic_to_dict

logger = logging.getLogger(__name__)


class SampleRateAnalysis(BaseModel):
    median: float
    mean: float
    std: float
    unit: str

    def to_pandas(self) -> pd.Series:
        return pd.Series(pydantic_to_dict(self)).to_frame().T

    def __repr__(self) -> str:
        return f"Median: {self.median} | {self.mean} +- {self.std} [{self.unit}]"
    

    def _repr_html_(self) -> str:
        return f"""<div> 
        <p> <span style="font-weight:bold"> Median: </span> {self.median} [{self.unit}]  </p>  
        <p> <span style="font-weight:bold">  Mean +- Std: </span> {self.mean:.3f} +- {self.std:.3f} [{self.unit}] </p>
        </div>
    """


def sample_rate(ds: AbstractPDMDataset, unit: str = "s") -> np.ndarray:
    """Obtain an array of time difference between two consecutive samples

    If the index it's a timestamp, the time difference will be converted to the provided unit

    Parameters:
        ds: The dataset
        unit: Unit to convert the timestamps differences

    Returns:
        Array of time differences

    """
    time_diff : List[float ]= []
    for life in ds:
        diff = np.diff(life.index.values)
        diff = diff[diff <= np.median(diff)]
        if pd.api.types.is_timedelta64_ns_dtype(diff.dtype):
            diff = diff / np.timedelta64(1, unit)
        time_diff.extend(diff)
    return np.array(time_diff)



def sample_rate_summary(
    ds: AbstractPDMDataset, unit: str = "s"
) -> SampleRateAnalysis:
    """
    Obtain the mean, median and standard deviation of the sample rate of the dataset

    Parameters:
        ds: The dataset
        unit: Unit to convert the time differences

    Returns:
        A SampleRateAnalysis with the following information: Mean sample rate, Std sample rate, Mode sample rate
    """
    sr = sample_rate(ds, unit)
    return SampleRateAnalysis(
        mean=np.mean(sr),
        std=np.std(sr),
        median=np.median(sr),
        unit=unit
    )



"""Windowed iteration capabilites

Usually time series data is divided in a contigous periods
of a determined window size, with some arbitray overlapping.
How the time series is iterated constitutes a very important
aspect when building a PdM model.

This module provides utilities to iterate the dataset



"""
from abc import abstractmethod
from enum import Enum
import logging
from signal import signal
from typing import Any, Callable, Optional, Tuple, Union

import numpy as np
import pandas as pd
from ceruleo.dataset.transformed import TransformedDataset
from ceruleo.iterators.sample_weight import (
    AbstractSampleWeights,
    NotWeighted,
    SampleWeight,
)
from ceruleo.iterators.shufflers import AbstractShuffler, NotShuffled
from tqdm.auto import tqdm
import functools


logger = logging.getLogger(__name__)


def seq_to_seq_signal_generator(
    signal_X: np.ndarray,
    signal_Y: np.ndarray,
    i: int,
    window_size: int,
    output_size: int = 1,
    right_closed: bool = True,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Generator for sequence to sequence models

    Parameters:
        signal_X: The input signal
        signal_Y: The output signal
        i: Current index
        window_size: Window size
        output_size: Output sequence length
        right_closed: Weather the last input of the window is included or not

    Returns:
        A Tuple with the Input and output sequences

    """
    initial = max(i - window_size + 1, 0)
    is_df = isinstance(signal_X, pd.DataFrame)
    if is_df:
        signal_X_1 = signal_X.iloc[initial : i + (1 if right_closed else 0), :]
    else:
        signal_X_1 = signal_X[initial : i + (1 if right_closed else 0), :]

    if is_df:
        signal_y_1 = signal_Y.iloc[initial : i + (1 if right_closed else 0), :]
    else:
        signal_y_1 = signal_Y[initial : i + (1 if right_closed else 0), :]

    return (signal_X_1, signal_y_1)


def windowed_signal_generator(
    data: pd.DataFrame,
    target: pd.DataFrame,
    i: int,
    window_size: int,
    output_size: int = 1,
    right_closed: bool = True,
) -> Tuple[np.ndarray, float]:
    """
    Return a lookback window and the value to predict.

    Parameters:
        data: Matrix of size (life_length, n_features) with the information of the life
        target: Target feature of size (life_length)
        i: Position of the value to predict
        window_size: Size of the lookback window
        output_size: Number of points of the target
        right_closed: Wether the last sample of the window should be included or not


    Returns:
        A tuple containing the lookback window and the value to predict.

    """
    initial = max(i - window_size + 1, 0)
    is_df = isinstance(data, pd.DataFrame)
    if is_df:
        signal_X_1 = data.iloc[initial : i + (1 if right_closed else 0), :].values
    else:
        signal_X_1 = data[initial : i + (1 if right_closed else 0), :]

    if len(target.shape) == 1:
        if is_df:
            signal_y_1 = target.iloc[i : min(i + output_size, target.shape[0])].values
        else:
            signal_y_1 = target[i : min(i + output_size, target.shape[0])]

        if signal_y_1.shape[0] < output_size:
            padding = np.zeros(output_size - signal_y_1.shape[0])
            signal_y_1 = np.hstack((signal_y_1, padding))
        signal_y_1 = np.expand_dims(signal_y_1, axis=1)
    else:
        if is_df:
            signal_y_1 = target.iloc[
                i : min(i + output_size, target.shape[0]), :
            ].values
        else:
            signal_y_1 = target[i : min(i + output_size, target.shape[0]), :]

        if signal_y_1.shape[0] < output_size:
            padding = np.zeros(
                ((output_size - signal_y_1.shape[0]), signal_y_1.shape[1])
            )
            signal_y_1 = np.concatenate((signal_y_1, padding), axis=0)

    if signal_X_1.shape[0] < window_size:
        signal_X_1 = np.vstack(
            (
                np.zeros((window_size - signal_X_1.shape[0], signal_X_1.shape[1])),
                signal_X_1,
            )
        )

    return (signal_X_1, signal_y_1)


class IterationType(Enum):
    """
    Iteration type

    Possible values are:

    - SEQ_TO_SEQ = 1: The seq to seq iterator will return as a target a window of a same size as the input aligned with it
    - FORECAST = 2: The forecast iterator produces as target the values of the Y transformers
        that start where the X data ends.
    """

    SEQ_TO_SEQ = 1
    FORECAST = 2


def valid_sample(
    padding: int,
    window_size: int,
    current_sample: int,
    y: int,
):
    if not padding:
        return current_sample >= window_size - 1
    else:
        return True


class RelativePosition:
    """
    Relative position selector base class

    The relative position selectors allow specifying the iteration starts and end relative to the beginning or the end of the run-to-cycle failure
    """

    def __init__(self, i: int):
        self.i = i

    @abstractmethod
    def get(self, time_series_length: int):
        raise NotImplementedError


class RelativeToEnd(RelativePosition):
    """
    Specify positions relative to the end of the run-to-failure cycle

    Example:
        An iterator that iterates each run-to-failure cycle starting in the last 500 samples of each cycle.
        ```
        iterator = WindowedDatasetIterator(
                transformed_ds,
                window_size=3,
                step=1,
                start_index=RelativeToEnd(500),
                horizon=1)
        ```
    """

    def get(self, time_series_length: int):
        return max(time_series_length - self.i, 0)


class RelativeToStart(RelativePosition):
    """
    Specify positions relative to the start of the run-to-failure cycle

    Example:
        An iterator that iterate each run-to-failure cycle skipping the first 200 samples of each cycle.
        ```
        iterator = WindowedDatasetIterator(
                transformed_ds,
                window_size=3,
                step=1,
                start_index=RelativeToStart(25),
                horizon=1)
        ```
    """

    def get(self, time_series_length: int):
        return self.i


class WindowedDatasetIterator:
    """
    Iterate a dataset using windows

    Parameters:
        dataset: The transformed dataset
        window_size: Size of the lookback window
        step: Separation between two consecutive size. If step == window_size there is no overlapping between two consecutive windows
        horizon: Horizon to be predicted. If this value is 3, for each window, 3 elements of the target are expected to be predicted
        shuffler: How the data should be shuffled
        sample_weight: Which are the sample weight for each sample
        right_closed: Weather the last point of the window should be included or not
        padding: Wether to pad elements if the samples are not enough to fill the window. Usually this happens at the beginning of the window
        iteration_type: Specify if the underlying model it's a forecasting model in which a scalar is predicted, or a sequence to sequence model similar to an autoencoder
        start_index: Initial index of each run-tu-failure cycle
        end_index: Final index of each run-to-failure cycle
        valid_sample: A callable that returns weather a sample is valid or not
        last_point: Weather to add the last point
    """

    def __init__(
        self,
        dataset: TransformedDataset,
        window_size: int,
        step: int = 1,
        horizon: int = 1,
        shuffler: AbstractShuffler = NotShuffled(),
        sample_weight: SampleWeight = NotWeighted(),
        right_closed: bool = True,
        padding: bool = False,
        iteration_type: IterationType = IterationType.FORECAST,
        start_index: Union[int, RelativePosition] = 0,
        end_index: Optional[Union[int, RelativePosition]] = None,
        valid_sample: Callable[[int, int, int, int, int], bool] = valid_sample,
        last_point: bool = True,
    ):
        self.last_point = last_point
        if isinstance(start_index, int):
            start_index = RelativeToStart(start_index)
        self.start_index = start_index
        if end_index is None:
            end_index = RelativeToEnd(0)
        elif isinstance(end_index, int):
            end_index = RelativeToStart(end_index)
        self.end_index = end_index
        self.dataset = dataset
        self.shuffler = shuffler
        self.window_size = window_size
        self.step = step
        self.shuffler.initialize(self)
        self.iteration_type = iteration_type

        if self.iteration_type == IterationType.FORECAST:
            self.slicing_function = windowed_signal_generator
        else:
            self.slicing_function = seq_to_seq_signal_generator

        if not isinstance(sample_weight, AbstractSampleWeights) or not callable(
            sample_weight
        ):
            raise ValueError(
                "sample_weight should be an AbstractSampleWeights or a callable"
            )

        self.sample_weight = sample_weight

        self.i = 0
        self.horizon = horizon
        self.right_closed = right_closed
        self.length = None
        self.padding = padding
        self.valid_sample = functools.partial(
            valid_sample, self.padding, self.window_size
        )

    def __len__(self):
        """
        Return the length of the iterator

        If it was not iterated once, it will compute the length by iterating from the entire dataset
        """
        if self.length is None:
            self.length = sum(1 for _ in self)
            self.__iter__()
        return self.length

    def __iter__(self):
        self.i = 0
        self.shuffler.start(self)
        return self

    def __next__(self):
        life, timestamp = self.shuffler.next_element()
        X, y, metadata = self.dataset[life]
        is_df = isinstance(y, pd.DataFrame)
        if is_df:
            valid = self.valid_sample(timestamp, y.iloc[timestamp])
        else:
            valid = self.valid_sample(timestamp, y[timestamp])
        while not valid:
            life, timestamp = self.shuffler.next_element()
            X, y, metadata = self.dataset[life]
            if is_df:
                valid = self.valid_sample(timestamp, y.iloc[timestamp])
            else:
                valid = self.valid_sample(timestamp, y[timestamp])

        curr_X, curr_y = self.slicing_function(
            X, y, timestamp, self.window_size, self.horizon, self.right_closed
        )
        return curr_X, curr_y, [self.sample_weight(y, timestamp, metadata)]

    def get_data(
        self, flatten: bool = True, show_progress: bool = False
    ) -> Tuple[np.ndarray, np.array, np.array]:
        """
        Obtain data, target and sample weights as numpy arrays.

        Parameters:
            flatten: Wether to flatten data
            show_progress: Wether to show progress

        Returns:
            Data, target and sample weights
        """
        N_points = len(self)

        if flatten:
            dimension = self.window_size * self.n_features
            X = np.zeros((N_points, dimension), dtype=np.float32)
        else:
            X = np.zeros(
                (N_points, self.window_size, self.n_features), dtype=np.float32
            )
        if self.iteration_type == IterationType.FORECAST:
            y = np.zeros((N_points, self.horizon), dtype=np.float32)
        else:
            y = np.zeros((N_points, self.window_size, self.horizon), dtype=np.float32)
        sample_weight = np.zeros(N_points, dtype=np.float32)

        iterator = enumerate(self)
        if show_progress:
            iterator = tqdm(iterator, total=len(self))

        for i, (X_, y_, sample_weight_) in iterator:
            if flatten:
                X[i, :] = X_.flatten()
            else:
                X[i, :, :] = X_
            if self.iteration_type == IterationType.FORECAST:
                y[i, :] = y_.flatten()
            else:
                y[i, :, :] = y_
            sample_weight[i] = sample_weight_[0]
        return X, y, sample_weight

    @property
    def n_features(self) -> int:
        """Number of features of the transformed dataset. This is a helper method to obtain the transformed dataset information from the WindowedDatasetIterator

        Returns:
            Number of features of the transformed dataset
        """
        return self.dataset.transformer.n_features

    @property
    def shape(self) -> Tuple[int, int]:
        """
        Tuple containing (window_size, n_features)

        Returns:
            Tuple containing (window_size, n_features)
        """
        return (self.window_size, self.n_features)



import logging
from typing import Optional, Union

import numpy as np
import pandas as pd
from ceruleo.transformation import TransformerStep
from ceruleo.transformation.features.tdigest import TDigest

logger = logging.getLogger(__name__)
           

class PerColumnImputer(TransformerStep):
    """Impute the values of each column following a simple rule

    The imputing is made following this rule:
        * -np.inf -> min
        * np.inf -> max
        * nan -> median

    Parameters:
        name: Step name, by default None
        
    """

    def __init__(self, *, name: Optional[str] = None):
        super().__init__(name=name, prefer_partial_fit=False)
        self.data_min = None
        self.data_max = None
        self.data_median = None

    def partial_fit(self, X:pd.DataFrame, y=None) -> pd.DataFrame:
        """ 
        Fit the transformation incrementally

        Parameters:
            X: The input life
        
        """
        X = X.replace([np.inf, -np.inf], np.nan)
        col_to_max = X.max()
        col_to_min = X.max()
        col_to_median = X.median()
        if self.data_min is None:
            self.data_min = col_to_min
            self.data_max = col_to_max
            self.data_median = col_to_median
        else:
            self.data_min = pd.concat([self.data_min, col_to_min], axis=1).min(axis=1)
            self.data_max = pd.concat([self.data_max, col_to_max], axis=1).max(axis=1)
            self.data_median = pd.concat([self.data_max, col_to_median], axis=1).median(
                axis=1
            )
        self._remove_na()

    def _remove_na(self):
        self.data_max.fillna(0, inplace=True)
        self.data_min.fillna(0, inplace=True)
        self.data_median.fillna(0, inplace=True)

    def fit(self, X:pd.DataFrame, y=None) -> pd.DataFrame:
        """ 
        Fit the transformation 

        Parameters:
            X: The input life
        
        """
        X = X.replace([np.inf, -np.inf], np.nan)
        col_to_max = X.max()
        col_to_min = X.max()
        col_to_median = X.median()

        self.data_min = col_to_min
        self.data_max = col_to_max
        self.data_median = col_to_median

        self._remove_na()
        return self

    def transform(self, X:pd.DataFrame, y=None) -> pd.DataFrame:
        """ 
        Apply the transformation to the input life

        Parameters:
            X: The input life
        
        """
        X_new = X.copy()
        for c in X_new.columns:
            X_new[c].replace([np.inf], self.data_max[c], inplace=True)
            X_new[c].replace([-np.inf], self.data_min[c], inplace=True)
            X_new[c].replace([np.nan], self.data_median[c], inplace=True)
        return X_new

    def description(self) -> tuple:
        """ 
        Transformation's Description

        Returns:
            A tuple with the transformation name and the Max, Min and Median values for each feature. 

        """
        name = super().description()
        data = []
        for k in self.data_max.index:
            data.append(
                (
                    k,
                    {
                        "Max": self.data_max[k],
                        "Min": self.data_max[k],
                        "Median": self.data_max[k],
                    },
                )
            )
        return (name, data)


class NaNtoInf(TransformerStep):
    """Replace NaN for inf"""

    def transform(self, X: pd.DataFrame, y=None) -> pd.DataFrame:
        """
        Transform the input life replacing Nan for inf

        Parameters:
            X: Input Dataframe to be transformed
         
        Returns:
            A dataframe with she same index as the input with the NaN values replaced with inf
        """
        return X.replace([np.inf, -np.inf], np.nan)
        #It should be -> X.replace(np.nan,[np.inf, -np.inf])
        #return X.replace(np.nan,np.inf)


class MedianImputer(TransformerStep):
    """
    Impute missing values with the median value of the training set

    Parameters:
        name: The name of the step
        
    """

    def __init__(self, *, name: Optional[str] = None):
        super().__init__(name=name)
        self.tdigest_dict = None

    def fit(self, X:pd.DataFrame, y=None):
        """Compute the median value

        Parameters:
            X: The input life
        """
        self.median = X.median(axis=0).to_dict()
        return self

    def partial_fit(self, X:pd.DataFrame):
        """Compute the median value incrementally

         Parameters:
            X: The input life
        """
        if self.tdigest_dict is None:
            self.tdigest_dict = {c: TDigest() for c in X.columns}
        for c in X.columns:
            self.tdigest_dict[c].batch_update(X[c].values)

        self.median = {
            c: self.tdigest_dict[c].percentile(50) for c in self.tdigest_dict.keys()
        }

    def transform(self, X: pd.DataFrame, y=None) -> pd.DataFrame:
        """
        Return a new dataframe with the missing values replaced by the fitted median

        Parameters:
            X: The input life


        Returns:
            A new DataFrame with the same index as the input with the Na values replaced by the fitted median
        """
        return X.fillna(value=self.median)


class MeanImputer(TransformerStep):
    """Impute missing values with the mean value of the training set

    Parameters:
        name: The name of the step
        
    """

    def __init__(self, *, name: Optional[str] = None):
        super().__init__(name=name)
        self.sum = None

    def partial_fit(self, X:pd.DataFrame, y=None):
        """Compute the mean value incrementally

        Parameters:
            X: The input life  
        """
        if self.sum is None:
            self.sum = X.sum(axis=0)
            self.counts = X.shape[0]
        else:
            self.sum += X.sum(axis=0)
            self.counts += X.shape[0]
        self.mean = (self.sum / self.counts).to_dict()
        return self

    def fit(self, X:pd.DataFrame, y=None):
        """Compute the mean value 

        Parameters:
            X: The input life  
        """
        self.mean = X.mean(axis=0).to_dict()
        return self

    def transform(self, X: pd.DataFrame, y=None) -> pd.DataFrame:
        """Return a new dataframe with the missing values replaced by the fitted mean

        Parameters:
            X: The input life

        Returns:
            A new DataFrame with the same index as the input with the Na values replaced by the fitted mean
        """
        return X.fillna(value=self.mean)


class ApplyRollingImputer(TransformerStep):
    """
    Impute missing values using a function over a rolling window

    Parameters:
        window_size: Window size of the rolling window
        func: The function to call in each window
        
    """

    def __init__(self, *, window_size: int, func: callable, **kwargs):
        super().__init__(**kwargs)
        self.window_size = window_size
        self.function = func
        self.mean_value_list = []
        self.sum = None

    def partial_fit(self, X: pd.DataFrame):
        """
        Compute incrementally the mean value to use as default value to impute

        Parameters:
            X: The input life
        """
        if self.sum is None:
            self.sum = X.sum(axis=0)
            self.counts = X.shape[0]
        else:
            self.sum += X.sum(axis=0)
            self.counts += X.shape[0]
        self.default_value = (self.sum / self.counts).to_dict()
        return self

    def fit(self, X: pd.DataFrame):
        """
        Compute a default value in case there are not valid values in the rolling window

        Parameters:
            X: The input life
        """
        self.default_value = np.mean(X, axis=0)
        self.default_value[~np.isfinite(self.default_value)] = 0
        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Transform the input life

        Parameters:
            X: The input life to be transformed

        Returns:
            A new life with the same index as the input with the missing values replaced by the output of the function supplied
        """
        X = X.copy()
        
        row, features = np.where(~np.isfinite(X))
        min_limit = np.maximum(row - self.window_size, 0)
        max_limit = np.minimum(row + self.window_size, X.shape[0])
        for r, min_r, max_r, f in zip(row, min_limit, max_limit, features):
            X.iloc[r, f] = self.function(X.iloc[min_r:max_r, f].values)
            if ~np.isfinite(X.iloc[r, f]):
                X.iloc[r, f] = self.default_value[X.columns[f]]
        return X


class RollingMedianImputer(ApplyRollingImputer):
    """Impute missing values with the median value on a rolling window

    Parameters:
        window_size: Window size of the rolling window
    """

    def __init__(self, *, window_size: int, name:str=None):
        super().__init__(window_size, func=np.median, name=name)


class RollingMeanImputer(ApplyRollingImputer):
    """Impute missing values with the mean value on a rolling window

    Parameters:
        window_size: Window size of the rolling window
            
    """

    def __init__(self, *, window_size: int, name:str=None):
        super().__init__(window_size=window_size, func=np.mean, name=name)


class ForwardFillImputer(TransformerStep):
    """Impute forward filling the values"""

    def transform(self, X:pd.DataFrame) -> pd.DataFrame:
        """
        Transform the input life

        Parameters:
            X: The input life to be transformed

        Returns:
            A new life with the same index as the input with the missing values replaced by the value in the succesive timestamp 
        """
        if not isinstance(X, pd.DataFrame):
            raise ValueError("Input array must be a data frame")
        return X.ffill()


class BackwardFillImputer(TransformerStep):
    """Impute forward filling the values"""

    def transform(self, X:pd.DataFrame) -> pd.DataFrame:
        """
        Transform the input life

        Parameters:
            X: The input life to be transformed

        Returns:
            A new life with the same index as the input with the missing values replaced by the value in the previous timestamp 
        """
        if not isinstance(X, pd.DataFrame):
            raise ValueError("Input array must be a data frame")
        return X.bfill()


class FillImputer(TransformerStep):
    def __init__(self, *, value: float, name: Optional[str] = None):
        super().__init__(name=name)
        self.value = value

    """Impute substituting the missing values with a value specified in the input 
    """
    def transform(self, X:pd.DataFrame) -> pd.DataFrame:
        """ 
        Transform the input life

        Parameters:
            X: The input life to be transformed

        Returns:
            A new life with the same index as the input with the missing values replaced by the value specified in the input 
        """
        return X.fillna(value=self.value)



"""
In some cases It's useful to cast the datatype of subset of columns

"""
import logging
from typing import List, Optional

import pandas as pd
from ceruleo.transformation import TransformerStep

logger = logging.getLogger(__name__)


class CastTo(TransformerStep):
    """Cast to a given datatype

    Example:
        step = CastTo(type='float32')

    Parameters:
        type: Data Type to cast to
        name: Name of the step, by default None

    """

    def __init__(self, *, type: str, name: Optional[str] = None):
        super().__init__(name=name)
        self.type = type

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Cast to a given datatype

        Parameters:
            X: DataFrame to transform

        Returns:
            Transformed DataFrame
        """
        return X.astype(self.type)


class ToDateTime(TransformerStep):
    """Cast to a datetime

    Example:
        step = ToDateTime()

    Parameters:
        name: Name of the step, by default None

    """

    def __init__(
        self,
        *,
        units: str = "s",
        name: Optional[str] = None,
        columns: Optional[List[any]] = None,
        index: bool = False,
        
    ):
        """Cast to a datetime

        Parameters
        ----------
        units : str, optional
            The units to transform, by default "s"
        name : Optional[str], optional
            Name of the step, by default None
        columns : Optional[List[any]], optional
            Columns to transform, by default None
        index : bool, optional
            Wether to transform the index or not, by default False

        Raises
        ------
        ValueError
            _description_
        ValueError
            _description_
        """
        super().__init__(name=name)
        if index is False and columns is None:
            raise ValueError("You must specify the columns to transform")
        if columns is not None and index is True:
            raise ValueError("You can't specify columns and index at the same time")
        self.index = index
        self.columns = columns
        self.units = units

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Cast to a datetime

        Parameters:
            X: DataFrame to transform

        Returns:
            Transformed DataFrame
        """
        X = X.copy()
        if self.index:
            X.index = pd.to_datetime(X.index, unit=self.units)
        else:
            X[self.columns] = X[self.columns].apply(pd.to_datetime)
        return X


from typing import List, Optional
from ceruleo.transformation.functional.transformerstep import TransformerStep
from copy import copy
from pyinform import active_info, block_entropy, entropy_rate
import pandas as pd
import numpy as np


ENTROPY_MEASURES_VALID_STATS = [
    "local_active_information",
    "local_block_entropy",
    "local_entropy_rate",
]


class LocalEntropyMeasures(TransformerStep):
    """Compute diverse entropy measures

    For each feature present in the life a number of feature will be computed for each time stamp

    The possible features are:

    - Local Active Information
    - Local Block Entropy
    - Local Entropy Rate


    Parameters:
        min_points: The minimun number of points of the expanding window, by default 2
        to_compute: List of the features to compute, by default None. Valid values are: 'local_active_information', 'local_block_entropy', 'local_entropy_rate'
        name: Name of the step, by default None
    """

    def __init__(
        self, window: int = 2, to_compute: List[str] = None, name: Optional[str] = None
    ):
        super().__init__(name=name)
        self.window = window

        if to_compute is None:
            self.to_compute = copy(ENTROPY_MEASURES_VALID_STATS)
        else:
            for f in to_compute:
                if f not in ENTROPY_MEASURES_VALID_STATS:
                    raise ValueError(
                        f"Invalid feature to compute {f}. Valids are {ENTROPY_MEASURES_VALID_STATS}"
                    )
            self.to_compute = to_compute

    def partial_fit(self, X, y=None):
        return self

    def fit(self, X, y=None):
        return self


    def _local_active_information(self, s: pd.Series):
        return active_info(s.values, self.window, local=True)

    def _local_block_entropy(self, s: pd.Series):
        return block_entropy(s.values, self.window, local=True)

    def _local_entropy_rate(self, s: pd.Series):
        return entropy_rate(s.values, self.window, local=True)

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:

        """ 
        Return a new dataframe with Entropy features computed for each feature in the input

        Parameters:
            X: Input life
        
        Returns:
            A new DataFrame with the at maximum three times the number of columns as the input since three Entropy features are computed for each feature in the input
        """

        X_new_n_columns = len(X.columns) * len(self.to_compute)
        i = 0

        columns = np.empty((X_new_n_columns,), dtype=object)
        for c in X.columns:
            for stats in self.to_compute:
                columns[i] = f"{c}_{stats}"
                i += 1
        
        


        data = np.empty((len(X.index), len(columns),))
        data[:] = np.nan
        X_new = pd.DataFrame(data, index=X.index, columns=columns)

        for c in X.columns:
            for stats in self.to_compute:
                data = np.squeeze(getattr(self, f"_{stats}")(X[c]))
                X_new.loc[:, f"{c}_{stats}"].iloc[-len(data):] = data
        X_new[np.isinf(X_new) | np.isnan(X_new)] = np.nan
        return X_new
